# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__doc__ = (
    "A sub-package to better organize the package's constants."
)


import types
import typing


class SimpleNamespace(types.SimpleNamespace):
    def __getitem__(self, name: str) -> typing.Any:
        try:
            return self.__dict__[name]
        except KeyError:
            return getattr(self, name)

    def __iter__(self) -> typing.Hashable:
        yield from self.__dict__

    def keys(self) -> typing.Hashable:
        yield from self.__dict__


from .misc import *
from .datasets import *
from .passcrypt import *
from .chunky2048 import *  # Chunky2048 consts get namespace priority


__all__ = [
    "misc",
    "datasets",
    "passcrypt",
    "chunky2048",
    *(n for n in globals() if n[0].isupper()),
]


misc = SimpleNamespace(**misc)
datasets = SimpleNamespace(**datasets)
passcrypt = SimpleNamespace(**passcrypt)
chunky2048 = SimpleNamespace(**chunky2048)




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = [
    "DomainKDF",
    "Ed25519",
    "PackageSigner",
    "PackageVerifier",
    "Passcrypt",
    "X25519",
    "agenerate_key",
    "amnemonic",
    "generate_key",
    "mnemonic",
]


__doc__ = (
    "A collection of high-level tools for creating & managing symmetric"
    " & 25519 elliptic curve asymmetric keys."
)


import json
import math
import cryptography
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PublicKey,
    X25519PrivateKey,
)
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PublicKey,
    Ed25519PrivateKey,
)
from pathlib import Path
from secrets import token_bytes
from functools import wraps, partial
from hashlib import sha384, sha3_256, sha3_512, shake_128, shake_256
from .__constants import *
from ._exceptions import *
from ._typing import Typing as t
from ._containers import PasscryptHash, PasscryptSettings
from ._containers import PackageSignerFiles, PackageSignerScope
from .asynchs import sleep, asleep, Processes
from .commons import Slots, OpenNamespace
from .commons import make_module
from .gentools import Comprende, comprehension
from .gentools import data, adata
from .gentools import bytes_range, abytes_range
from .generics import Domains, Clock
from .generics import xi_mix, axi_mix
from .generics import encode_key, aencode_key
from .generics import hash_bytes, ahash_bytes
from .generics import int_as_bytes, bytes_as_int
from .generics import canonical_pack, acanonical_pack
from .generics import bytes_are_equal, abytes_are_equal
from .randoms import csprng, acsprng
from .randoms import generate_key, agenerate_key
from .randoms import generate_salt, agenerate_salt
from .ciphers import KeyAADBundle
from .ciphers import bytes_keys, abytes_keys


ns_clock = Clock(NANOSECONDS)
day_clock = Clock(DAYS, epoch=0)


class DomainKDF:
    """
    Creates objects able to derive domain & payload-specific keyed
    hashes. Payload updates are automatically canonicalized.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp import DomainKDF


    kdf = DomainKDF(b"ecdhe", session.transcript, key=session.key)

    auth_key = kdf.sha3_512(context=b"auth-key")
    encryption_key = kdf.sha3_512(context=b"encryption-key")

    """

    __slots__ = ("_domain", "_key", "_payload")

    _key_size: int = 0  # for subclasses which use an XOF for keygen
    _key_type: callable = sha3_512
    _type: callable = shake_256

    _KEY_TYPE_BLOCKSIZE: int = _key_type().block_size
    _TYPE_BLOCKSIZE: int = _type().block_size

    def _initialize_payload(self, payload: t.Iterable[bytes]) -> None:
        """
        Canonically encodes the key & first batch of input data to
        prepare a hashing object which will process all additional
        payload.
        """
        payload_key = encode_key(self._key, self._TYPE_BLOCKSIZE)
        payload = (
            canonical_pack(*payload, blocksize=self._TYPE_BLOCKSIZE)
            if payload
            else b""
        )
        self._payload = self._type(payload_key + payload)

    def __init__(
        self, domain: bytes, *payload: t.Iterable[bytes], key: bytes
    ) -> "self":
        """
        Initializes a ``domain``-specific key & a keyed-mac object to
        incorporate an arbitrary amount of ``payload`` data, which is
        canonically encoded by the class, in key derivation.
        """
        self._domain = domain
        self._key = key
        self.update_key(key)
        self._initialize_payload(payload)

    def copy(self) -> "cls":
        """
        Copies the instance state into a new object which can be updated
        separately in differing contexts.
        """
        kdf = self.__class__.__new__(self.__class__)
        kdf._key = self._key
        kdf._domain = self._domain
        kdf._payload = self._payload.copy()
        return kdf

    async def aupdate(self, *payload: t.Iterable[bytes]) -> "self":
        """
        Canonically update the payload object with the additional
        arguments passed in as ``payload``. This allows large amounts of
        data to be used for key derivation without a large in-memory
        cost. Also, mitigates canonicalization attacks. Update calls,
        input data & the order of both must match exactly to create the
        same KDF state with two different objects.
        """
        if not payload:
            raise Issue.value_must("update payload", "not be empty")
        self._payload.update(
            await acanonical_pack(*payload, blocksize=self._TYPE_BLOCKSIZE)
        )
        return self

    def update(self, *payload: t.Iterable[bytes]) -> "self":
        """
        Canonically update the payload object with the additional
        arguments passed in as ``payload``. This allows large amounts of
        data to be used for key derivation without a large in-memory
        cost. Also, mitigates canonicalization attacks. Update calls,
        input data & the order of both must match exactly to create the
        same KDF state with two different objects.
        """
        if not payload:
            raise Issue.value_must("update payload", "not be empty")
        self._payload.update(
            canonical_pack(*payload, blocksize=self._TYPE_BLOCKSIZE)
        )
        return self

    async def aupdate_key(self, entropy: bytes) -> "self":
        """
        Derive's a new instance key from the its domain, current key &
        the provided ``entropy``.
        """
        self._key = await ahash_bytes(
            Domains.KDF,
            self._domain,
            entropy,
            key=self._key + entropy,
            hasher=self._key_type,
            size=self._key_size,
        )
        return self

    def update_key(self, entropy: bytes) -> "self":
        """
        Derive's a new instance key from the its domain, current key &
        the provided ``entropy``.
        """
        self._key = hash_bytes(
            Domains.KDF,
            self._domain,
            entropy,
            key=self._key + entropy,
            hasher=self._key_type,
            size=self._key_size,
        )
        return self

    async def asha3_256(self, *, context: bytes = b"") -> bytes:
        """
        Return the keyed sha3_256 hash of the instance's state.
        """
        return await ahash_bytes(
            self._domain,
            context,
            self._payload.digest(SHA3_256_BLOCKSIZE),
            key=self._key + context,
            hasher=sha3_256,
        )

    def sha3_256(self, *, context: bytes = b"") -> bytes:
        """
        Return the keyed sha3_256 hash of the instance's state.
        """
        return hash_bytes(
            self._domain,
            context,
            self._payload.digest(SHA3_256_BLOCKSIZE),
            key=self._key + context,
            hasher=sha3_256,
        )

    async def asha3_512(self, *, context: bytes = b"") -> bytes:
        """
        Return the keyed sha3_512 hash of the instance's state.
        """
        return await ahash_bytes(
            self._domain,
            context,
            self._payload.digest(SHA3_512_BLOCKSIZE),
            key=self._key + context,
            hasher=sha3_512,
        )

    def sha3_512(self, *, context: bytes = b"") -> bytes:
        """
        Return the keyed sha3_512 hash of the instance's state.
        """
        return hash_bytes(
            self._domain,
            context,
            self._payload.digest(SHA3_512_BLOCKSIZE),
            key=self._key + context,
            hasher=sha3_512,
        )

    async def ashake_128(self, size: int, *, context: bytes = b"") -> bytes:
        """
        Return the keyed shake_128 hash of the instance's state.
        """
        return await ahash_bytes(
            self._domain,
            context,
            self._payload.digest(SHAKE_128_BLOCKSIZE),
            size=size,
            key=self._key + context,
            hasher=shake_128,
        )

    def shake_128(self, size: int, *, context: bytes = b"") -> bytes:
        """
        Return the keyed shake_128 hash of the instance's state.
        """
        return hash_bytes(
            self._domain,
            context,
            self._payload.digest(SHAKE_128_BLOCKSIZE),
            size=size,
            key=self._key + context,
            hasher=shake_128,
        )

    async def ashake_256(self, size: int, *, context: bytes = b"") -> bytes:
        """
        Return the keyed shake_256 hash of the instance's state.
        """
        return await ahash_bytes(
            self._domain,
            context,
            self._payload.digest(SHAKE_256_BLOCKSIZE),
            size=size,
            key=self._key + context,
            hasher=shake_256,
        )

    def shake_256(self, size: int, *, context: bytes = b"") -> bytes:
        """
        Return the keyed shake_256 hash of the instance's state.
        """
        return hash_bytes(
            self._domain,
            context,
            self._payload.digest(SHAKE_256_BLOCKSIZE),
            size=size,
            key=self._key + context,
            hasher=shake_256,
        )


async def akeyed_choices(
    choices: t.Sequence[t.Any],
    selection_size: int,
    *,
    domain: bytes = b"",
    key: bytes,
) -> t.AsyncGenerator[None, t.Any]:
    """
    Makes ``selection_size`` number of selections from an indexable
    sequence of ``choices`` using subkeys derived from a provided
    ``domain`` & ``key``. Yields each selection one at a time.
    """
    total_choices = len(choices)
    key = await DomainKDF(
        domain,
        int_as_bytes(total_choices),
        int_as_bytes(selection_size),
        key=key,
    ).ashake_256(16 * selection_size, context=Domains.PRNG)
    async for index in adata.root(key, size=16):
        yield choices[bytes_as_int(index) % total_choices]


def keyed_choices(
    choices: t.Sequence[t.Any],
    selection_size: int,
    *,
    domain: bytes = b"",
    key: bytes,
) -> t.Generator[None, t.Any, None]:
    """
    Makes ``selection_size`` number of selections from an indexable
    sequence of ``choices`` using subkeys derived from a provided
    ``domain`` & ``key``. Yields each selection one at a time.
    """
    total_choices = len(choices)
    key = DomainKDF(
        domain,
        int_as_bytes(total_choices),
        int_as_bytes(selection_size),
        key=key,
    ).shake_256(16 * selection_size, context=Domains.PRNG)
    for index in data.root(key, size=16):
        yield choices[bytes_as_int(index) % total_choices]


async def amnemonic(
    passphrase: bytes = b"",
    size: int = 8,
    *,
    salt: t.Optional[bytes] = b"",
    words: t.Optional[t.Sequence[t.Any]] = None,
    **passcrypt_settings: t.PasscryptKWsNew,
) -> t.List[bytes]:
    """
    Creates list of ``size`` number of words for a mnemonic key from a
    user ``passphrase`` & an optional ``salt``. If no ``passphrase`` is
    supplied, then a random value is used to derive a unique mnemonic.
    The ``words`` used for the mnemonic can be passed, but by default
    are a word-list of 2048 unique, all lowercase english words.
    """
    domain = Domains.MNEMONIC
    words = words if words else WORD_LIST
    salt = (salt + domain) if salt else domain
    if passphrase:
        key = await Passcrypt.anew(passphrase, salt, **passcrypt_settings)
    elif passcrypt_settings:
        raise Issue.unused_parameters(
            "passcrypt_settings", "generating a random mnemonic key"
        )
    else:
        key = await acsprng()
    return [
        word
        async for word in akeyed_choices(
            words, size, domain=domain, key=key
        )
    ]


def mnemonic(
    passphrase: bytes = b"",
    size: int = 8,
    *,
    salt: t.Optional[bytes] = b"",
    words: t.Optional[t.Sequence[t.Any]] = None,
    **passcrypt_settings: t.PasscryptKWsNew,
) -> t.List[bytes]:
    """
    Creates list of ``size`` number of words for a mnemonic key from a
    user ``passphrase`` & an optional ``salt``. If no ``passphrase`` is
    supplied, then a random value is used to derive a unique mnemonic.
    The ``words`` used for the mnemonic can be passed, but by default
    are a word-list of 2048 unique, all lowercase english words.
    """
    domain = Domains.MNEMONIC
    words = words if words else WORD_LIST
    salt = (salt + domain) if salt else domain
    if passphrase:
        key = Passcrypt.new(passphrase, salt, **passcrypt_settings)
    elif passcrypt_settings:
        raise Issue.unused_parameters(
            "passcrypt_settings", "generating a random mnemonic key"
        )
    else:
        key = csprng()
    return [*keyed_choices(words, size, domain=domain, key=key)]


class PasscryptSession(Slots):
    """
    Hanldes the initialization of running the `Passcrypt` hashing
    algorithm with sets of given user parameters.
    """

    __slots__ = (
        "passphrase",
        "salt",
        "aad",
        "mb",
        "cpu",
        "cores",
        "tag_size",
        "row_size",
        "rows",
        "total_size",
        "ram",
        "proof",
    )

    vars().update({var: passcrypt[var] for var in passcrypt})

    _PASSCRYPT_KDF_SALT: bytes = Domains.encode_constant(
        b"passcrypt_kdf_salt", size=SHAKE_128_BLOCKSIZE
    )

    _new_passcrypt_proof_kdf: callable = shake_128(_PASSCRYPT_KDF_SALT).copy

    specification: OpenNamespace = OpenNamespace(
        is_passphrase=lambda passphrase: (
            len(passphrase) >= passcrypt.MIN_PASSPHRASE_BYTES
            and passphrase.__class__ is bytes
        ),
        is_salt=lambda salt: (
            len(salt)
            in range(passcrypt.MIN_SALT_SIZE, passcrypt.MAX_SALT_SIZE + 1)
            and salt.__class__ is bytes
        ),
        is_aad=lambda aad: aad.__class__ is bytes,
        is_mb=lambda mb: (
            mb in range(passcrypt.MIN_MB, passcrypt.MAX_MB + 1)
            and mb.__class__ is int
        ),
        is_cpu=lambda cpu: (
            cpu in range(passcrypt.MIN_CPU, passcrypt.MAX_CPU + 1)
            and cpu.__class__ is int
        ),
        is_cores=lambda cores: (
            cores in range(passcrypt.MIN_CORES, passcrypt.MAX_CORES + 1)
            and cores.__class__ is int
        ),
        is_tag_size=lambda tag_size: (
            tag_size >= passcrypt.MIN_TAG_SIZE and tag_size.__class__ is int
        ),
        is_salt_size=lambda salt_size: (
            salt_size
            in range(passcrypt.MIN_SALT_SIZE, passcrypt.MAX_SALT_SIZE + 1)
            and salt_size.__class__ is int
        ),
    )

    @classmethod
    def _validate_inputs(
        cls, passphrase: bytes, salt: bytes, aad: bytes
    ) -> None:
        """
        Makes sure ``passphrase``, ``salt`` & ``aad`` are to the
        ``Passcrypt`` specification. Throws `ValueError` or `TypeError`
        accordingly.
        """
        if not cls.specification.is_passphrase(passphrase):
            raise PasscryptIssue.improper_passphrase(Metadata(passphrase))
        elif not cls.specification.is_salt(salt):
            raise PasscryptIssue.improper_salt(Metadata(salt))
        elif not cls.specification.is_aad(aad):
            raise PasscryptIssue.improper_aad()

    @classmethod
    def _validate_settings(
        cls,
        mb: int,
        cpu: int,
        cores: int,
        tag_size: int,
        salt_size: int = passcrypt.MIN_SALT_SIZE,
    ) -> None:
        """
        Ensures the values ``mb``, ``cpu``, ``cores`` & ``tag_size``
        passed into this package's Argon2i-like, passphrase-based key
        derivation function are within acceptable bounds & types. The
        ``salt_size`` can be validated as well. Since it's not used in
        all interfaces, it's optional here.
        """
        if not cls.specification.is_mb(mb):
            raise PasscryptIssue.invalid_mb(mb)
        elif not cls.specification.is_cpu(cpu):
            raise PasscryptIssue.invalid_cpu(cpu)
        elif not cls.specification.is_cores(cores):
            raise PasscryptIssue.invalid_cores(cores)
        elif not cls.specification.is_tag_size(tag_size):
            raise PasscryptIssue.invalid_tag_size(tag_size)
        elif not cls.specification.is_salt_size(salt_size):
            raise PasscryptIssue.invalid_salt_size(salt_size)

    def __init__(
        self,
        passphrase: bytes,
        salt: bytes,
        *,
        aad: bytes,
        mb: int,
        cpu: int,
        cores: int,
        tag_size: int,
    ) -> "self":
        """
        Efficiently stores user parameters.
        """
        self._validate_inputs(passphrase, salt, aad=aad)
        self._validate_settings(mb, cpu, cores, tag_size)
        self.passphrase = passphrase
        self.salt = salt
        self.aad = aad
        self.mb = mb
        self.cpu = cpu
        self.cores = cores
        self.tag_size = tag_size

    def __iter__(
        self,
    ) -> t.Generator[None, t.Union[bytearray, t.Callable, int], None]:
        """
        Dumps the set of relevant parameters & function pointers for the
        `Passcrypt` worker. Gives the session a cleaner interface.
        """
        yield from (
            self.ram,
            self.proof.update,
            self.proof.digest,
            self.row_size,
            self.total_size,
        )

    def _hash_session_parameters(self) -> bytes:
        """
        Returns a 336-byte hash of the session's canonically encoded
        parameters.
        """
        return hash_bytes(
            Domains.PASSCRYPT,
            self.salt,
            self.aad,
            int_as_bytes(self.mb),
            int_as_bytes(self.cpu),
            int_as_bytes(self.cores),
            int_as_bytes(self.tag_size),
            key=self.passphrase + self.salt + self.aad,
            hasher=shake_128,
            size=336,
            pad=self.PASSCRYPT_PAD,
        )

    def prepare_session(self) -> "self":
        """
        Canonically hash the parameters to the function & calculate the
        dimensionality of the cache from the given settings.
        """
        rounds = max(
            (1, self.cpu // self.CPU_TO_DIGEST_PAIRS_PER_ROW_RATIO)
        )
        self.rows = math.ceil((B_TO_MB_RATIO * self.mb) / (336 * rounds))
        self.row_size = 336 * rounds
        self.total_size = self.row_size * self.rows
        parameters = self._hash_session_parameters()
        self.proof = self._new_passcrypt_proof_kdf()
        self.proof.update(parameters)
        return self

    def allocate_ram(self) -> "self":
        """
        Builds a virtual 2d memory cache out of a 1d bytearray to do
        efficient & in-place memory overwrites of segments of the cache
        with new proofs-of-work as the `Passcrypt` algorithm runs to
        completion.

        The bytearray is traversed to simulate the dimensionality of a
        columns=2*rounds, rows=ceil((1024*1024*mb) / (2*168*rounds)),
        2d array, where the unit measure for the width of one column is
        168-bytes (one digest from the `shake_128` `proof` object), &
        rounds=max([1, cpu // 2]).

        This procedure is designed to build the initial cache as fast as
        possible using the C implementation of `hashlib.shake_128` to
        better equalize the execution time between users & their
        adversaries. Quickly building the initial cache to the full size
        of the desired `mb` memory cost is also intended to reduce the
        inefficiencies of doing any resizing of the cache once the
        algorithm begins. This too is the main motivating factor for the
        size of each row being an equal multiple of 336, as it allows
        cache traversal & insertions without needing to plan separately
        for how to treat insertions once the end of a row is reached.
        """
        self.ram = bytearray()
        size = self.total_size
        max_size = (B_TO_MB_RATIO * 512) - 1  # 512MiB, max digest size
        while size > max_size:                # of shake_128 in python
            self.ram.extend(self.proof.digest(max_size))
            self.proof.update(self.ram[-168:])
            size -= max_size
        if size:
            self.ram.extend(self.proof.digest(size))
            self.proof.update(self.ram[-168:])
        return self


class Passcrypt:
    """
    This class is used to implement an Argon2i-like passphrase-based
    key derivation function that's designed to be resistant to cache-
    timing side-channel attacks & time-memory trade-offs.

    It uses a passphrase-keyed scanning function which sequentially
    passes over unique memory caches requiring a tunable amount of
    difficulty, which is designed here to be very intuitive.

    This scheme is secret independent with regard to how it chooses to
    pass over memory. Through proofs of work & memory, it ensures an
    attacker attempting to crack a passphrase hash cannot complete the
    algorithm substantially faster by storing more memory than what's
    already necessary, or with substantially less memory, by dropping
    cache entries, without drastically increasing the computational
    cost.

    The algorithm initializes all of the columns for the cache using a
    single `shake_128` object after being fed the passphrase, salt, aad
    & all of the parameters. The number of columns is computed
    dynamically to reach the specified memory cost considering that each
    row will hold 2 * max([1, cpu // 2]) digests of 168-bytes. This
    allows the cache to be efficiently allocated up front, benefiting
    further by not needing to resize the memory cache throughout the
    running of the algorithm.

    The sequential passes involve a current row index, the index of the
    row which is the reflection of the first across the cache, & a
    current offset into a row which are multiples of 336 (two digests).
    The index & reflection pointers interleave each other, hashing rows
    with the same object as they scan, & overwriting the 168-byte digest
    / piece of cache at the offset they're pointing to after each hash.

     _____________________________________
    |                                     |
    |    Algorithm Diagram: Side-View     |
    |_____________________________________|

           ___________________ # of rows ___________________
          |                                                 |
          |              initial memory cache               |
          |  row  # of columns == 2 * max([1, cpu // 2])    |
          |   |   # of rows == ⌈1024*1024*mb/168*columns⌉   |
          v   v                                             v
    column|---'-----------------------------------------'---| the initial cache
    column|---'-----------------------------------------'---| of size ~`mb` is
    column|---'-----------------------------------------'---| built very quickly
    column|---'-----------------------------------------'---| using SHAKE-128.
    column|---'-----------------------------------------'---| each (row, column)
    column|---'-----------------------------------------'---| coordinate holds
    column|---'-----------------------------------------'---| one element of
    column|---'-----------------------------------------'---| 168-bytes.
                                                        ^
                                                        |
                           reflection                  row
                          <-   |
          |--------------------'-------'--------------------| each row is
          |--------------------'-------'--------------------| hashed then has
          |--------------------'-------'--------------------| a new 168-byte
          |--------------------'-------'--------------------| digest overwrite
          |--------------------'-------'--------------------| the current pointer
          |--------------------'-------'--------------------| in an alternating
          |--------------------Xxxxxxxx'xxxxxxxxxxxxxxxxxxxx| sequence, first at
          |oooooooooooooooooooo'oooooooO--------------------| the index, then at
                                       |   ->                 its reflection.
                                     index


          |--'-------------------------------------------'--| this continues
          |--'-------------------------------------------'--| until the entire
          |--'-------------------------------------------Xxx| cache has been
          |ooO-------------------------------------------'--| overwritten.
          |xx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'xx| a single `shake_128`
          |oo'ooooooooooooooooooooooooooooooooooooooooooo'oo| object (H) is used
          |xx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx'xx| to do all of the
          |oo'ooooooooooooooooooooooooooooooooooooooooooo'oo| hashing.
             |   ->                                 <-   |
           index                                     reflection


          |xxxxxxxxxxx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx| finally, the whole
          |ooooooooooo'ooooooooooooooooooooooooooooooooooooo| cache is quickly
          |xxxxxxxxxxx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx| hashed `cpu` + 2
          |ooooooooooo'ooooooooooooooooooooooooooooooooooooo| number of times.
          |Fxxxxxxxxxx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx| after each pass an
          |foooooooooo'ooooooooooooooooooooooooooooooooooooo| 84-byte digest is
          |fxxxxxxxxxx'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx| inserted into the
          |foooooooooo'ooooooooooooooooooooooooooooooooooooo| cache, ruling out
                      |   ->                                  hashing state cycles.
                      | hash cpu + 2 # of times               Then a `tag_size`-
                      v                                       byte tag is output.
                  H(cache)

          tag = H.digest(tag_size)

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp import Passcrypt, Domains
    from getpass import getpass


    Passcrypt.PEPPER = bytes.fromhex(app.application_pepper)

    # registration ->

    un = getpass("username: ").encode() + Domains.USERNAME
    pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

    pw_hash = Passcrypt.hash_passphrase(pw, aad=un, mb=128)


    # a login attempt ->

    un = getpass("username: ").encode() + Domains.USERNAME
    pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

    try:
        Passcrypt.verify(pw_hash, pw, aad=un, ttl=24 * 3600)
    except Passcrypt.InvalidPassphrase as auth_fail:
        app.post_mortem(error=auth_fail)
    except Passcrypt.TimestampExpired as expired_hash:
        # 24-hour registration expired
        app.post_mortem(error=expired_hash)

    """

    __slots__ = ()

    vars().update({f"_{var}": passcrypt[var] for var in passcrypt})

    # An operator of a passphrase database may add a static secret value
    # to the class, referred to as a `pepper`. That value can be set in
    # this variable & will then augment all hashes produced by the class
    # with that additional secret entropy. The operator is in charge of
    # storing this value securely so it can be reused when the program
    # restarts. This value SHOULD NOT be stored in the same database
    # where the hashes are stored.
    PEPPER: bytes = b""

    TimestampExpired = TimestampExpired
    InvalidPassphrase = InvalidPassphrase
    ImproperPassphrase = ImproperPassphrase

    def __new__(
        cls,
        mb: int = _DEFAULT_MB,
        cpu: int = _DEFAULT_CPU,
        cores: int = _DEFAULT_CORES,
        tag_size: int = _DEFAULT_SCHEMA_TAG_SIZE,
        salt_size: int = _DEFAULT_SCHEMA_SALT_SIZE,
    ) -> "PasscryptInstance":
        """
        Stores user-defined settings within an object so that they can
        automatically be passed into mirrored calls to the class'
        methods, simulating the behavior of instance methods.
        """
        return PasscryptInstance(
            mb=mb,
            cpu=cpu,
            cores=cores,
            tag_size=tag_size,
            salt_size=salt_size,
        )

    @staticmethod
    def _work_memory_prover(session: PasscryptSession) -> bytes:
        """
        Returns the digest of a keyed scanning function. It sequentially
        passes over a memory cache with an intuitive & tunable amount of
        difficulty. This scheme is secret independent with regard to how
        it chooses to pass over memory.

        Through proofs of work & memory, it ensures an attacker
        attempting to crack a passphrase hash cannot complete the
        algorithm substantially faster by storing more memory than
        what's already necessary, or with substantially less memory, by
        dropping cache entries, without drastically increasing the
        computational cost.
        """
        ram, update, digest, row_size, total_size = session
        assert total_size == len(ram)
        column_start_indexes = range(0, row_size, 336)
        row_start_indexes = [*range(0, total_size, row_size)]
        for column_start in column_start_indexes:
            for row_start in row_start_indexes:
                index = row_start + column_start
                ref_row_start = -row_start - row_size
                reflection = ref_row_start + column_start + 168
                ref_end = reflection + 168
                ref_end = ref_end if ref_end < 0 else None

                update(ram[row_start : row_start + row_size])
                ram[index : index + 168] = digest(168)

                update(ram[ref_row_start : ref_row_start + row_size])
                ram[reflection:ref_end] = digest(168)
        for iteration in range(session.cpu + 2):
            seek = 168 * iteration
            ram[seek : seek + 84] = digest(84)
            update(ram)
        return digest(session.tag_size)

    @classmethod
    def _passcrypt(cls, session: PasscryptSession) -> bytes:
        """
        This method implements an Argon2i-like passphrase-based key
        derivation function that's designed to be resistant to cache-
        timing side-channel attacks & time-memory trade-offs.
        """
        session.prepare_session().allocate_ram()
        return cls._work_memory_prover(session)

    @classmethod
    async def anew(
        cls,
        passphrase: bytes,
        salt: bytes,
        *,
        aad: bytes = _DEFAULT_AAD,
        mb: int = _DEFAULT_MB,
        cpu: int = _DEFAULT_CPU,
        cores: int = _DEFAULT_CORES,
        tag_size: int = _DEFAULT_TAG_SIZE,
    ) -> bytes:
        """
        Returns just the ``tag_size``-byte hash of the ``passphrase``
        when processed with the given ``salt``, ``aad`` & difficulty
        settings.

        ``passphrase``: A 12-byte or greater entropic value that
                contains the user's desired entropy & cryptographic
                strength. A passphrase is ideally a compilation of
                words, symbols & numbers which are easy for the user to
                remember, but difficult for anyone else to guess even if
                they know precise details about who the user is & their
                characteristics / tendencies.

        ``salt``: An ephemeral, uniform 8-byte or greater entropic
                value. SHOULD BE RANDOMLY GENERATED, its main purpose
                is to be unpredictable.

        ``aad``: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different ``aad``.

        ``mb``: The int number of Mibibytes (MiB) the user desires for
                this run of the algorithm to cost to compute.

        ``cpu``: The int number of iterations over the entire memory
                cache that are executed for this run of the algorithm.
                The size & computational cost of each row of the memory
                cache is also linearly proportional to this parameter.

        ``cores``: The int number of separate processes that will be
                pooled to compute this run of the algorithm.

        ``tag_size``: The int length of the hash that will be output at
                the end of this run of the algorithm.

        NOTICE: The passcrypt algorithm can be highly memory intensive.
        These resources may not be freed up, & often are not, because of
        python quirks around memory management. To force the release of
        these resources, we run the function in another process which
        guarantees the release.
        """
        if not PasscryptSession.specification.is_cores(cores):
            raise PasscryptIssue.invalid_cores(cores)
        try:
            sessions = []
            total_mb = int_as_bytes(mb)
            core_mb = math.ceil(mb / cores)
            async for core in abytes_range.root(cores):
                core_aad = canonical_pack(cls.PEPPER, aad, core, total_mb)
                session = PasscryptSession(
                    passphrase,
                    salt,
                    aad=core_aad,
                    mb=core_mb,
                    cpu=cpu,
                    cores=cores,
                    tag_size=tag_size,
                )
                kw = dict(session=session, probe_frequency=0.001)
                sessions.append(
                    await Processes.asubmit(cls._passcrypt, **kw)
                )
            for core, session in enumerate(sessions):
                sessions[core] = await session.aresult()
            return await axi_mix(b"".join(sessions), size=tag_size)
        except Exception as error:
            if Processes._pool._broken:
                Processes.reset_pool()
                raise Issue.broken_pool_restarted() from error
            raise error

    @classmethod
    def new(
        cls,
        passphrase: bytes,
        salt: bytes,
        *,
        aad: bytes = _DEFAULT_AAD,
        mb: int = _DEFAULT_MB,
        cpu: int = _DEFAULT_CPU,
        cores: int = _DEFAULT_CORES,
        tag_size: int = _DEFAULT_TAG_SIZE,
    ) -> bytes:
        """
        Returns just the ``tag_size``-byte hash of the ``passphrase``
        when processed with the given ``salt``, ``aad`` & difficulty
        settings.

        ``passphrase``: A 12-byte or greater entropic value that
                contains the user's desired entropy & cryptographic
                strength. A passphrase is ideally a compilation of
                words, symbols & numbers which are easy for the user to
                remember, but difficult for anyone else to guess even if
                they know precise details about who the user is & their
                characteristics / tendencies.

        ``salt``: An ephemeral, uniform 8-byte or greater entropic
                value. SHOULD BE RANDOMLY GENERATED, its main purpose
                is to be unpredictable.

        ``aad``: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different ``aad``.

        ``mb``: The int number of Mibibytes (MiB) the user desires for
                this run of the algorithm to cost to compute.

        ``cpu``: The int number of iterations over the entire memory
                cache that are executed for this run of the algorithm.
                The size & computational cost of each row of the memory
                cache is also linearly proportional to this parameter.

        ``cores``: The int number of separate processes that will be
                pooled to compute this run of the algorithm.

        ``tag_size``: The int length of the hash that will be output at
                the end of this run of the algorithm.

        NOTICE: The passcrypt algorithm can be highly memory intensive.
        These resources may not be freed up, & often are not, because of
        python quirks around memory management. To force the release of
        these resources, we run the function in another process which
        guarantees the release.
        """
        if not PasscryptSession.specification.is_cores(cores):
            raise PasscryptIssue.invalid_cores(cores)
        try:
            sessions = []
            total_mb = int_as_bytes(mb)
            core_mb = math.ceil(mb / cores)
            for core in bytes_range.root(cores):
                core_aad = canonical_pack(cls.PEPPER, aad, core, total_mb)
                session = PasscryptSession(
                    passphrase,
                    salt,
                    aad=core_aad,
                    mb=core_mb,
                    cpu=cpu,
                    cores=cores,
                    tag_size=tag_size,
                )
                kw = dict(session=session, probe_frequency=0.001)
                sessions.append(Processes.submit(cls._passcrypt, **kw))
            for core, session in enumerate(sessions):
                sessions[core] = session.result()
            return xi_mix(b"".join(sessions), size=tag_size)
        except Exception as error:
            if Processes._pool._broken:
                Processes.reset_pool()
                raise Issue.broken_pool_restarted() from error
            raise error

    @classmethod
    async def ahash_passphrase(
        cls,
        passphrase: bytes,
        *,
        aad: bytes = _DEFAULT_AAD,
        mb: int = _DEFAULT_MB,
        cpu: int = _DEFAULT_CPU,
        cores: int = _DEFAULT_CORES,
        tag_size: int = _DEFAULT_SCHEMA_TAG_SIZE,
        salt_size: int = _DEFAULT_SCHEMA_SALT_SIZE,
    ) -> bytes:
        """
        Returns the passcrypt difficulty settings, salt & hash of the
        ``passphrase`` in a single raw bytes sequence for convenient
        storage. The salt here is automatically generated.

        ``passphrase``: A 12-byte or greater entropic value that
                contains the user's desired entropy & cryptographic
                strength. A passphrase is ideally a compilation of
                words, symbols & numbers which are easy for the user to
                remember, but difficult for anyone else to guess even if
                they know precise details about who the user is & their
                characteristics / tendencies.

        ``aad``: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different ``aad``.

        ``mb``: The int number of Mibibytes (MiB) the user desires for
                this run of the algorithm to cost to compute.

        ``cpu``: The int number of iterations over the entire memory
                cache that are executed for this run of the algorithm.
                The size & computational cost of each row of the memory
                cache is also linearly proportional to this parameter.

        ``cores``: The int number of separate processes that will be
                pooled to compute this run of the algorithm.

        ``tag_size``: The int length of the hash that will be output at
                the end of this run of the algorithm.

        ``salt_size``: The int length of the salt that will be generated
                & attached to the hash for this run of the algorithm.

         _____________________________________
        |                                     |
        |   Format Diagram: Passcrypt Hash    |
        |_____________________________________|
         ______________________________________________________________
        |                                    |                         |
        |               Header               |          Body           |
        |-------|-------|------|------|------|--------|----------------|
        | time  |  mb   | cpu  | cores| Slen |  salt  |      tag       |
        |  8-   |  3-   |  1-  |  1-  |  1-  |  Slen- |      >=16-     |
        | bytes | bytes | byte | byte | byte |  bytes |     bytes      |
        |_______|_______|______|______|______|________|________________|
        |                                                              |
        |                          >=34-bytes                          |
        |______________________________________________________________|

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import Passcrypt, Domains
        from getpass import getpass


        Passcrypt.PEPPER = bytes.fromhex(getpass("hexidecimal pepper: "))

        # registration ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

        pw_hash = await Passcrypt.ahash_passphrase(pw, aad=un, mb=128)


        # a login attempt ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

        try:
            await Passcrypt.averify(pw_hash, pw, aad=un, ttl=24 * 3600)
        except Passcrypt.InvalidPassphrase as auth_fail:
            app.post_mortem(error=auth_fail)
        except Passcrypt.TimestampExpired as expired_hash:
            # 24-hour registration expired
            app.post_mortem(error=expired_hash)

        """
        timestamp = await ns_clock.amake_timestamp()
        salt = token_bytes(salt_size)
        tag = await cls.anew(
            passphrase,
            salt,
            aad=timestamp + aad,
            mb=mb,
            cpu=cpu,
            cores=cores,
            tag_size=tag_size,
        )
        return PasscryptHash(
            timestamp=timestamp,
            mb=mb,
            cpu=cpu,
            cores=cores,
            salt=salt,
            tag=tag,
        ).export_hash()

    @classmethod
    def hash_passphrase(
        cls,
        passphrase: bytes,
        *,
        aad: bytes = _DEFAULT_AAD,
        mb: int = _DEFAULT_MB,
        cpu: int = _DEFAULT_CPU,
        cores: int = _DEFAULT_CORES,
        tag_size: int = _DEFAULT_SCHEMA_TAG_SIZE,
        salt_size: int = _DEFAULT_SCHEMA_SALT_SIZE,
    ) -> bytes:
        """
        Returns the passcrypt difficulty settings, salt & hash of the
        ``passphrase`` in a single raw bytes sequence for convenient
        storage. The salt here is automatically generated.

        ``passphrase``: A 12-byte or greater entropic value that
                contains the user's desired entropy & cryptographic
                strength. A passphrase is ideally a compilation of
                words, symbols & numbers which are easy for the user to
                remember, but difficult for anyone else to guess even if
                they know precise details about who the user is & their
                characteristics / tendencies.

        ``aad``: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different ``aad``.

        ``mb``: The int number of Mibibytes (MiB) the user desires for
                this run of the algorithm to cost to compute.

        ``cpu``: The int number of iterations over the entire memory
                cache that are executed for this run of the algorithm.
                The size & computational cost of each row of the memory
                cache is also linearly proportional to this parameter.

        ``cores``: The int number of separate processes that will be
                pooled to compute this run of the algorithm.

        ``tag_size``: The int length of the hash that will be output at
                the end of this run of the algorithm.

        ``salt_size``: The int length of the salt that will be generated
                & attached to the hash for this run of the algorithm.

         _____________________________________
        |                                     |
        |   Format Diagram: Passcrypt Hash    |
        |_____________________________________|
         ______________________________________________________________
        |                                    |                         |
        |               Header               |          Body           |
        |-------|-------|------|------|------|--------|----------------|
        | time  |  mb   | cpu  | cores| Slen |  salt  |      tag       |
        |  8-   |  3-   |  1-  |  1-  |  1-  |  Slen- |      >=16-     |
        | bytes | bytes | byte | byte | byte |  bytes |     bytes      |
        |_______|_______|______|______|______|________|________________|
        |                                                              |
        |                          >=34-bytes                          |
        |______________________________________________________________|

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import Passcrypt, Domains
        from getpass import getpass


        Passcrypt.PEPPER = bytes.fromhex(getpass("hexidecimal pepper: "))

        # registration ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

        pw_hash = Passcrypt.hash_passphrase(pw, aad=un, mb=128)


        # a login attempt ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

        try:
            Passcrypt.verify(pw_hash, pw, aad=un, ttl=24 * 3600)
        except Passcrypt.InvalidPassphrase as auth_fail:
            app.post_mortem(error=auth_fail)
        except Passcrypt.TimestampExpired as expired_hash:
            # 24-hour registration expired
            app.post_mortem(error=expired_hash)

        """
        timestamp = ns_clock.make_timestamp()
        salt = token_bytes(salt_size)
        tag = cls.new(
            passphrase,
            salt,
            aad=timestamp + aad,
            mb=mb,
            cpu=cpu,
            cores=cores,
            tag_size=tag_size,
        )
        return PasscryptHash(
            timestamp=timestamp,
            mb=mb,
            cpu=cpu,
            cores=cores,
            salt=salt,
            tag=tag,
        ).export_hash()

    @classmethod
    async def averify(
        cls,
        composed_passcrypt_hash: bytes,
        passphrase: bytes,
        *,
        aad: bytes = _DEFAULT_AAD,
        ttl: int = DEFAULT_TTL,
        mb_allowed: range = _MB_RESOURCE_SAFETY_RANGE,
        cpu_allowed: range = _CPU_RESOURCE_SAFETY_RANGE,
        cores_allowed: range = _CORES_RESOURCE_SAFETY_RANGE,
    ) -> None:
        """
        Verifies that a supplied ``passphrase`` was indeed used to build
        the ``composed_passcrypt_hash``.

        Runs the passcrypt algorithm on the ``passphrase`` with the
        parameters specified in the ``composed_passcrypt_hash`` value's
        attached metadata. If the result doesn't match the hash in
        ``composed_passcrypt_hash`` then `Passcrypt.InvalidPassphrase`
        is raised. The ``composed_passcrypt_hash`` passed into this
        method must be raw bytes.

        ``aad``: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different ``aad``.

        ``ttl``: An amount of seconds which dictate the allowable age of
                a ``composed_passcrypt_hash``. The associated timestamp,
                which is attached to the hash, helps ensure the tag is
                unique by separating each tag created across time into
                distinct domains.

        ``mb_allowed``: A `builtins.range` object which includes all
                allowable values for the `mb` (Mibibyte) resource cost.
                Raises `ResourceWarning` if the `mb` specified in the
                provided hash falls outside of that range.

        ``cpu_allowed``: A `builtins.range` object which includes all
                allowable values for the `cpu` resource cost. Raises
                `ResourceWarning` if the `cpu` specified in the provided
                hash falls outside of that range.

        ``cores_allowed``: A `builtins.range` object which includes all
                allowable values for the `cores` resource cost. Raises
                `ResourceWarning` if the `cores` specified in the
                provided hash falls outside of that range.

         _____________________________________
        |                                     |
        |   Format Diagram: Passcrypt Hash    |
        |_____________________________________|
         ______________________________________________________________
        |                                    |                         |
        |               Header               |          Body           |
        |-------|-------|------|------|------|--------|----------------|
        | time  |  mb   | cpu  | cores| Slen |  salt  |      tag       |
        |  8-   |  3-   |  1-  |  1-  |  1-  |  Slen- |      >=16-     |
        | bytes | bytes | byte | byte | byte |  bytes |     bytes      |
        |_______|_______|______|______|______|________|________________|
        |                                                              |
        |                          >=34-bytes                          |
        |______________________________________________________________|

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import Passcrypt, Domains
        from getpass import getpass


        Passcrypt.PEPPER = bytes.fromhex(getpass("hexidecimal pepper: "))

        # registration ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

        pw_hash = await Passcrypt.ahash_passphrase(pw, aad=un, mb=128)


        # a login attempt ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

        try:
            await Passcrypt.averify(pw_hash, pw, aad=un, ttl=24 * 3600)
        except Passcrypt.InvalidPassphrase as auth_fail:
            app.post_mortem(error=auth_fail)
        except Passcrypt.TimestampExpired as expired_hash:
            # 24-hour registration expired
            app.post_mortem(error=expired_hash)

        """
        parts = PasscryptHash().import_hash(composed_passcrypt_hash)
        await ns_clock.atest_timestamp(parts.timestamp, ttl * NS_TO_S_RATIO)
        parts.in_allowed_ranges(mb_allowed, cpu_allowed, cores_allowed)
        untrusted_hash = await cls.anew(
            passphrase,
            parts.salt,
            aad=parts.timestamp + aad,
            mb=parts.mb,
            cpu=parts.cpu,
            cores=parts.cores,
            tag_size=parts.tag_size,
        )
        if not bytes_are_equal(untrusted_hash, parts.tag):
            raise PasscryptIssue.verification_failed()

    @classmethod
    def verify(
        cls,
        composed_passcrypt_hash: bytes,
        passphrase: bytes,
        *,
        aad: bytes = _DEFAULT_AAD,
        ttl: int = DEFAULT_TTL,
        mb_allowed: range = _MB_RESOURCE_SAFETY_RANGE,
        cpu_allowed: range = _CPU_RESOURCE_SAFETY_RANGE,
        cores_allowed: range = _CORES_RESOURCE_SAFETY_RANGE,
    ) -> None:
        """
        Verifies that a supplied ``passphrase`` was indeed used to build
        the ``composed_passcrypt_hash``.

        Runs the passcrypt algorithm on the ``passphrase`` with the
        parameters specified in the ``composed_passcrypt_hash`` value's
        attached metadata. If the result doesn't match the hash in
        ``composed_passcrypt_hash`` then `Passcrypt.InvalidPassphrase`
        is raised. The ``composed_passcrypt_hash`` passed into this
        method must be raw bytes.

        ``aad``: An arbitrary bytes value a user decides to categorize
                the hash with. It is authenticated as associated data &
                safely permutes hashes created with different ``aad``.

        ``ttl``: An amount of seconds which dictate the allowable age of
                a ``composed_passcrypt_hash``. The associated timestamp,
                which is attached to the hash, helps ensure the tag is
                unique by separating each tag created across time into
                distinct domains.

        ``mb_allowed``: A `builtins.range` object which includes all
                allowable values for the `mb` (Mibibyte) resource cost.
                Raises `ResourceWarning` if the `mb` specified in the
                provided hash falls outside of that range.

        ``cpu_allowed``: A `builtins.range` object which includes all
                allowable values for the `cpu` resource cost. Raises
                `ResourceWarning` if the `cpu` specified in the provided
                hash falls outside of that range.

        ``cores_allowed``: A `builtins.range` object which includes all
                allowable values for the `cores` resource cost. Raises
                `ResourceWarning` if the `cores` specified in the
                provided hash falls outside of that range.

         _____________________________________
        |                                     |
        |   Format Diagram: Passcrypt Hash    |
        |_____________________________________|
         ______________________________________________________________
        |                                    |                         |
        |               Header               |          Body           |
        |-------|-------|------|------|------|--------|----------------|
        | time  |  mb   | cpu  | cores| Slen |  salt  |      tag       |
        |  8-   |  3-   |  1-  |  1-  |  1-  |  Slen- |      >=16-     |
        | bytes | bytes | byte | byte | byte |  bytes |     bytes      |
        |_______|_______|______|______|______|________|________________|
        |                                                              |
        |                          >=34-bytes                          |
        |______________________________________________________________|

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import Passcrypt, Domains
        from getpass import getpass


        Passcrypt.PEPPER = bytes.fromhex(getpass("hexidecimal pepper: "))

        # registration ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE


        pw_hash = Passcrypt.hash_passphrase(pw, aad=un, mb=128)

        # a login attempt ->

        un = getpass("username: ").encode() + Domains.USERNAME
        pw = getpass("passphrase: ").encode() + Domains.PASSPHRASE

        try:
            Passcrypt.verify(pw_hash, pw, aad=un, ttl=24 * 3600)
        except Passcrypt.InvalidPassphrase as auth_fail:
            app.post_mortem(error=auth_fail)
        except Passcrypt.TimestampExpired as expired_hash:
            # 24-hour registration expired
            app.post_mortem(error=expired_hash)

        """
        parts = PasscryptHash().import_hash(composed_passcrypt_hash)
        ns_clock.test_timestamp(parts.timestamp, ttl * NS_TO_S_RATIO)
        parts.in_allowed_ranges(mb_allowed, cpu_allowed, cores_allowed)
        untrusted_hash = cls.new(
            passphrase,
            parts.salt,
            aad=parts.timestamp + aad,
            mb=parts.mb,
            cpu=parts.cpu,
            cores=parts.cores,
            tag_size=parts.tag_size,
        )
        if not bytes_are_equal(untrusted_hash, parts.tag):
            raise PasscryptIssue.verification_failed()


class PasscryptInstance:
    """
    Gives the user objects which mirrors calls to `Passcrypt` methods
    with automated passing of instance settings.
    """

    __slots__ = ("_settings",)

    TimestampExpired = TimestampExpired
    InvalidPassphrase = InvalidPassphrase
    ImproperPassphrase = ImproperPassphrase

    def __init__(
        self, mb: int, cpu: int, cores: int, tag_size: int, salt_size: int
    ) -> "self":
        """
        Stores user-defined settings so they can automatically be passed
        into `Passcrypt` methods when they are called.
        """
        self._settings = PasscryptSettings(
            mb=mb,
            cpu=cpu,
            cores=cores,
            tag_size=tag_size,
            salt_size=salt_size,
        )
        PasscryptSession._validate_settings(**self._settings)

    @wraps(Passcrypt.ahash_passphrase.__func__)
    async def ahash_passphrase(
        self,
        *args: t.Iterable[bytes],
        **kwargs: t.Dict[str, t.Union[bytes, int]],
    ) -> bytes:
        """
        Forwards calls to the `Passcrypt.ahash_passphrase` method with
        the instance settings & the specified ``args`` & ``kwargs``.
        """
        return await Passcrypt.ahash_passphrase(
            *args, **{**self._settings, **kwargs}
        )

    @wraps(Passcrypt.hash_passphrase.__func__)
    def hash_passphrase(
        self,
        *args: t.Iterable[bytes],
        **kwargs: t.Dict[str, t.Union[bytes, int]],
    ) -> bytes:
        """
        Forwards calls to the `Passcrypt.hash_passphrase` method with
        the instance settings & the specified ``args`` & ``kwargs``.
        """
        return Passcrypt.hash_passphrase(
            *args, **{**self._settings, **kwargs}
        )

    @wraps(Passcrypt.averify.__func__)
    async def averify(
        self,
        *args: t.Iterable[bytes],
        **kwargs: t.Dict[str, t.Union[bytes, int, range]],
    ) -> None:
        """
        Forwards calls to the `Passcrypt.averify` method with the
        specified ``args`` & ``kwargs``.
        """
        await Passcrypt.averify(*args, **kwargs)

    @wraps(Passcrypt.verify.__func__)
    def verify(
        self,
        *args: t.Iterable[bytes],
        **kwargs: t.Dict[str, t.Union[bytes, int, range]],
    ) -> None:
        """
        Forwards calls to the `Passcrypt.verify` method with the
        specified ``args`` & ``kwargs``.
        """
        Passcrypt.verify(*args, **kwargs)


class Curve25519:
    """
    Contains a collection of class methods & values that simplify the
    usage of the cryptography library, as well as pointers to values in
    the cryptography library.
    """

    __slots__ = ()

    _PUBLIC_BYTES_ENUM = {
        "encoding": serialization.Encoding.Raw,
        "format": serialization.PublicFormat.Raw,
    }
    _PRIVATE_BYTES_ENUM = {
        "encoding": serialization.Encoding.Raw,
        "format": serialization.PrivateFormat.Raw,
        "encryption_algorithm": serialization.NoEncryption(),
    }

    X25519PublicKey = X25519PublicKey
    X25519PrivateKey = X25519PrivateKey
    Ed25519PublicKey = Ed25519PublicKey
    Ed25519PrivateKey = Ed25519PrivateKey

    cryptography = cryptography
    exceptions = cryptography.exceptions
    hazmat = cryptography.hazmat
    serialization = serialization

    @staticmethod
    async def aed25519_key() -> Ed25519PrivateKey:
        """
        Returns an ``Ed25519PrivateKey`` from the cryptography package
        used to make elliptic curve signatures of data.
        """
        await asleep()
        return Ed25519PrivateKey.generate()

    @staticmethod
    def ed25519_key() -> Ed25519PrivateKey:
        """
        Returns an ``Ed25519PrivateKey`` from the cryptography package
        used to make elliptic curve signatures of data.
        """
        return Ed25519PrivateKey.generate()

    @staticmethod
    async def ax25519_key() -> X25519PrivateKey:
        """
        Returns a ``X25519PrivateKey`` from the cryptography package for
        use in an elliptic curve diffie-hellman exchange.
        """
        await asleep()
        return X25519PrivateKey.generate()

    @staticmethod
    def x25519_key() -> X25519PrivateKey:
        """
        Returns a ``X25519PrivateKey`` from the cryptography package for
        use in an elliptic curve diffie-hellman exchange.
        """
        return X25519PrivateKey.generate()

    @classmethod
    async def apublic_bytes(
        cls,
        key: t.Union[
            X25519PrivateKey,
            Ed25519PrivateKey,
            X25519PublicKey,
            Ed25519PublicKey,
        ],
    ) -> bytes:
        """
        Returns the public key bytes of either an ``X25519PrivateKey``,
        ``X25519PublicKey``, ``Ed25519PublicKey`` or ``Ed25519PrivateKey``
        object from the cryptography package.
        """
        await asleep()
        key_types = (
            X25519PrivateKey,
            Ed25519PrivateKey,
            X25519PublicKey,
            Ed25519PublicKey,
        )
        if not issubclass(key.__class__, key_types):
            raise Issue.value_must_be_type("key", key_types)
        elif hasattr(key, "public_key"):
            public_key = key.public_key()
        else:
            public_key = key
        return public_key.public_bytes(**cls._PUBLIC_BYTES_ENUM)

    @classmethod
    def public_bytes(
        cls,
        key: t.Union[
            X25519PrivateKey,
            Ed25519PrivateKey,
            X25519PublicKey,
            Ed25519PublicKey,
        ],
    ) -> bytes:
        """
        Returns the public key bytes of either an ``X25519PrivateKey``,
        ``X25519PublicKey``, ``Ed25519PublicKey`` or ``Ed25519PrivateKey``
        object from the cryptography package.
        """
        key_types = (
            X25519PrivateKey,
            Ed25519PrivateKey,
            X25519PublicKey,
            Ed25519PublicKey,
        )
        if not issubclass(key.__class__, key_types):
            raise Issue.value_must_be_type("key", key_types)
        elif hasattr(key, "public_key"):
            public_key = key.public_key()
        else:
            public_key = key
        return public_key.public_bytes(**cls._PUBLIC_BYTES_ENUM)

    @classmethod
    async def asecret_bytes(
        cls, secret_key: t.Union[X25519PrivateKey, Ed25519PrivateKey]
    ) -> bytes:
        """
        Returns the secret key bytes of either an ``X25519PrivateKey``
        or ``Ed25519PrivateKey`` from the cryptography package.
        """
        await asleep()
        key_types = (X25519PrivateKey, Ed25519PrivateKey)
        if not issubclass(secret_key.__class__, key_types):
            raise Issue.value_must_be_type("secret_key", key_types)
        return secret_key.private_bytes(**cls._PRIVATE_BYTES_ENUM)

    @classmethod
    def secret_bytes(
        cls, secret_key: t.Union[X25519PrivateKey, Ed25519PrivateKey]
    ) -> bytes:
        """
        Returns the secret key bytes of either an ``X25519PrivateKey``
        or ``Ed25519PrivateKey`` from the cryptography package.
        """
        key_types = (X25519PrivateKey, Ed25519PrivateKey)
        if not issubclass(secret_key.__class__, key_types):
            raise Issue.value_must_be_type("secret_key", key_types)
        return secret_key.private_bytes(**cls._PRIVATE_BYTES_ENUM)

    @staticmethod
    async def aexchange(
        secret_key: X25519PrivateKey, public_key: bytes
    ) -> bytes:
        """
        Returns the shared key bytes derived from an elliptic curve key
        exchange with the user's ``secret_key`` key, & their communicating
        peer's ``public_key`` public key's bytes or hex value.
        """
        await asleep()
        return secret_key.exchange(
            X25519PublicKey.from_public_bytes(public_key)
        )

    @staticmethod
    def exchange(secret_key: X25519PrivateKey, public_key: bytes) -> bytes:
        """
        Returns the shared key bytes derived from an elliptic curve key
        exchange with the user's ``secret_key`` key, & their communicating
        peer's ``public_key`` public key's bytes or hex value.
        """
        return secret_key.exchange(
            X25519PublicKey.from_public_bytes(public_key)
        )


class Base25519:
    """
    Collects the shared functionality between the ``X25519`` & ``Ed25519``
    classes.
    """

    __slots__ = ("_public_key", "_secret_key")

    _Curve25519 = Curve25519

    _exceptions = Curve25519.exceptions

    PublicKey = None
    SecretKey = None

    @classmethod
    def _process_public_key(
        cls,
        public_key: t.Union[
            str,
            bytes,
            X25519PrivateKey,
            Ed25519PrivateKey,
            X25519PublicKey,
            Ed25519PublicKey,
        ],
    ) -> t.Union[X25519PublicKey, Ed25519PublicKey]:
        """
        Accepts a ``public_key`` in either hex, bytes, ``X25519PublicKey``,
        ``X25519PrivateKey``, ``Ed25519PublicKey`` or ``Ed25519PrivateKey``
        format. Returns an instantiaed public key associated with the
        subclass inhereting this method.
        """
        if not public_key:
            raise Issue.no_value_specified("public key")
        elif public_key.__class__ is str:
            public_key = bytes.fromhex(public_key)
        elif public_key.__class__ is not bytes:
            public_key = cls._Curve25519.public_bytes(public_key)
        return cls.PublicKey.from_public_bytes(public_key)

    @classmethod
    def _process_secret_key(
        cls,
        secret_key: t.Union[
            str, bytes, X25519PrivateKey, Ed25519PrivateKey
        ],
    ) -> t.Union[Ed25519PrivateKey, X25519PrivateKey]:
        """
        Accepts a ``secret_key`` in either hex, bytes, ``X25519PrivateKey``
        or ``Ed25519PrivateKey`` format. Returns an instantiaed secret
        key associated with the subclass inhereting this method.
        """
        if not secret_key:
            raise Issue.no_value_specified("secret key")
        elif secret_key.__class__ is str:
            secret_key = bytes.fromhex(secret_key)
        elif secret_key.__class__ is not bytes:
            secret_key = cls._Curve25519.secret_bytes(secret_key)
        return cls.SecretKey.from_private_bytes(secret_key)

    async def aimport_public_key(
        self,
        public_key: t.Union[
            str,
            bytes,
            X25519PrivateKey,
            Ed25519PrivateKey,
            X25519PublicKey,
            Ed25519PublicKey,
        ],
    ) -> "self":
        """
        Populates an instance from the received ``public_key`` that is
        of either hex, bytes, ``X25519PublicKey``, ``X25519PrivateKey``,
        ``Ed25519PublicKey`` or ``Ed25519PrivateKey`` type.
        """
        await asleep()
        if hasattr(self, "_public_key"):
            raise Issue.value_already_set("public key", "the instance")
        self._public_key = self._process_public_key(public_key)
        return self

    def import_public_key(
        self,
        public_key: t.Union[
            str,
            bytes,
            X25519PrivateKey,
            Ed25519PrivateKey,
            X25519PublicKey,
            Ed25519PublicKey,
        ],
    ) -> "self":
        """
        Populates an instance from the received ``public_key`` that is
        of either hex, bytes, ``X25519PublicKey``, ``X25519PrivateKey``,
        ``Ed25519PublicKey`` or ``Ed25519PrivateKey`` type.
        """
        if hasattr(self, "_public_key"):
            raise Issue.value_already_set("public key", "the instance")
        self._public_key = self._process_public_key(public_key)
        return self

    async def aimport_secret_key(
        self,
        secret_key: t.Union[
            str, bytes, X25519PrivateKey, Ed25519PrivateKey
        ],
    ) -> "self":
        """
        Populates an instance from the received ``secret_key`` that is
        of either hex, bytes, ``X25519PrivateKey`` or ``Ed25519PrivateKey``
        type.
        """
        await asleep()
        if hasattr(self, "_public_key") or hasattr(self, "_secret_key"):
            raise Issue.value_already_set(f"key", "the instance")
        self._secret_key = self._process_secret_key(secret_key)
        self._public_key = self.PublicKey.from_public_bytes(
            await self._Curve25519.apublic_bytes(self._secret_key)
        )
        return self

    def import_secret_key(
        self,
        secret_key: t.Union[
            str, bytes, X25519PrivateKey, Ed25519PrivateKey
        ],
    ) -> "self":
        """
        Populates an instance from the received ``secret_key`` that is
        of either hex, bytes, ``X25519PrivateKey`` or ``Ed25519PrivateKey``
        type.
        """
        if hasattr(self, "_public_key") or hasattr(self, "_secret_key"):
            raise Issue.value_already_set(f"key", "the instance")
        self._secret_key = self._process_secret_key(secret_key)
        self._public_key = self.PublicKey.from_public_bytes(
            self._Curve25519.public_bytes(self._secret_key)
        )
        return self

    @property
    def secret_key(self) -> t.Union[X25519PrivateKey, Ed25519PrivateKey]:
        """
        Returns the instantiated & populated SecretKey of the associated
        sublass inhereting this method.
        """
        return self._secret_key

    @property
    def public_key(self) -> t.Union[X25519PublicKey, Ed25519PublicKey]:
        """
        Returns the instantiated & populated PublicKey of the associated
        sublass inhereting this method.
        """
        return self._public_key

    @property
    def secret_bytes(self) -> bytes:
        """
        Returns the secret bytes of the instance's instantiated &
        populated SecretKey of the associated sublass inhereting this
        method.
        """
        return self._Curve25519.secret_bytes(self._secret_key)

    @property
    def public_bytes(self) -> bytes:
        """
        Returns the public bytes of the instance's instantiated &
        populated PublicKey of the associated sublass inhereting this
        method.
        """
        return self._Curve25519.public_bytes(self._public_key)

    def has_secret_key(self) -> bool:
        """
        Returns a boolean of whether the instance contains a secret key.
        """
        return hasattr(self, "_secret_key")

    def has_public_key(self) -> bool:
        """
        Returns a boolean of whether the instance contains a public key.
        """
        return hasattr(self, "_public_key")


class Ed25519(Base25519):
    """
    This class is used to create stateful objects that simplify usage of
    the cryptography library's ed25519 protocol.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp import Ed25519

    # In a land, long ago ->
    alices_key = Ed25519().generate()
    internet.send(alices_key.public_bytes)

    # Alice wants to sign a document so that Bob can prove she wrote it.
    # So, Alice sends her public key bytes of the key she wants to
    # associate with her identity, the document & the signature ->
    document = b"DesignDocument.cad"
    signed_document = alices_key.sign(document)
    message = {
        "document": document,
        "signature": signed_document,
        "public_key": alices_key.public_bytes,
    }
    internet.send(message)

    # In a land far away ->
    alices_message = internet.receive()

    # Bob sees the message from Alice! Bob already knows Alice's public
    # key & she has reason believe it is genuinely hers. She'll then
    # verify the signed document ->
    assert alices_message["public_key"] == alices_public_key
    alice_verifier = Ed25519().import_public_key(alices_public_key)
    alice_verifier.verify(
        alices_message["signature"], alices_message["document"]
    )
    internet.send(b"Beautiful work, Alice! Thanks ^u^")

    # The verification didn't throw an exception! So, Bob knows the file
    # was signed by Alice.
    """

    __slots__ = ("_public_key", "_secret_key")

    InvalidSignature = Base25519._exceptions.InvalidSignature
    PublicKey = Curve25519.Ed25519PublicKey
    SecretKey = Curve25519.Ed25519PrivateKey

    async def agenerate(self) -> "self":
        """
        Generates a new secret key used for signing bytes data &
        populates the instance with it & its associated public key. This
        method returns the instance for convenience in instantiating a
        stateful object with await Ed25519().agenerate().
        """
        key = await self._Curve25519.aed25519_key()
        await self.aimport_secret_key(key)
        return self

    def generate(self) -> "self":
        """
        Generates a new secret key used for signing bytes data &
        populates the instance with it & its associated public key. This
        method returns the instance for convenience in instantiating a
        stateful object with Ed25519().generate().
        """
        key = self._Curve25519.ed25519_key()
        self.import_secret_key(key)
        return self

    async def asign(self, data: bytes) -> bytes:
        """
        Signs some bytes ``data`` with the instance's secret key.
        """
        await asleep()
        return self.secret_key.sign(data)

    def sign(self, data: bytes) -> bytes:
        """
        Signs some bytes ``data`` with the instance's secret key.
        """
        return self.secret_key.sign(data)

    async def averify(
        self,
        signature: bytes,
        data: bytes,
        *,
        public_key: t.Union[None, str, bytes, Ed25519PublicKey] = None,
    ) -> None:
        """
        Receives a ``signature`` to verify data with the instance's
        public key. If the ``public_key`` keyword-only argument is
        used, then that key is used instead of the instance key to run
        the verification.
        """
        if public_key:
            await asleep()
            public_key = self._process_public_key(public_key)
        else:
            public_key = self.public_key
        await asleep()
        public_key.verify(signature, data)

    def verify(
        self,
        signature: bytes,
        data: bytes,
        *,
        public_key: t.Union[None, str, bytes, Ed25519PublicKey] = None,
    ) -> None:
        """
        Receives a ``signature`` to verify data with the instance's
        public key. If the ``public_key`` keyword-only argument is
        used, then that key is used instead of the instance key to run
        the verification.
        """
        if public_key:
            public_key = self._process_public_key(public_key)
        else:
            public_key = self.public_key
        public_key.verify(signature, data)


class X25519(Base25519):
    """
    This class is used to create stateful objects that simplify usage of
    the cryptography library's x25519 protocol.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    user_alice = X25519().generate()
    # Alice wants to create a shared key with Bob. So, Alice sends the
    # public key bytes of her new key to bob ->
    internet.send(user_alice.public_bytes.hex())

    # In a land far away ->
    alices_message = internet.receive()

    # Bob sees the message from Alice! So she creates a key to accept
    # the exchange & sends the public bytes back to Alice ->
    user_bob = await X25519().agenerate()
    shared_key = user_bob.exchange(alices_message)
    internet.send(user_bob.public_bytes.hex())

    # When Alice receives Bob's public key & finishes the exchange, they
    # will have a shared symmetric key to encrypt messages to one
    # another.
    bobs_response = internet.receive()
    shared_key = user_alice.exchange(bobs_response)

    This protocol is not secure against active adversaries that can
    manipulate the information while its in transit between Alice &
    Bob. Each public key should only be used once.
    """

    __slots__ = ("_public_key", "_secret_key")

    PublicKey = Curve25519.X25519PublicKey
    SecretKey = Curve25519.X25519PrivateKey

    async def agenerate(self) -> "self":
        """
        Generates a new secret key used for a single elliptic curve
        diffie-hellman exchange, or as an argument to one of the 3dh or
        2dh generators in X25519.protocols. This populates the instance
        with the secret key & its associated public key. This method
        returns the instance for convenience in instantiating a stateful
        object with await X25519().agenerate().
        """
        key = await self._Curve25519.ax25519_key()
        await self.aimport_secret_key(key)
        return self

    def generate(self) -> "self":
        """
        Generates a new secret key used for a single elliptic curve
        diffie-hellman exchange, or as an argument to one of the 3dh or
        2dh generators in X25519.protocols. This populates the instance
        with the secret key & its associated public key. This method
        returns the instance for convenience in instantiating a stateful
        object with await X25519().generate().
        """
        key = self._Curve25519.x25519_key()
        self.import_secret_key(key)
        return self

    async def aexchange(
        self, public_key: t.Union[X25519PublicKey, bytes, str]
    ) -> bytes:
        """
        Takes in a public key from a communicating party & uses the
        instance's secret key to do an elliptic curve diffie-hellman
        exchange & returns the resulting secret shared bytes.
        """
        await asleep()
        public_key = self._process_public_key(public_key)
        return await self._Curve25519.aexchange(
            self._secret_key,
            await self._Curve25519.apublic_bytes(public_key),
        )

    def exchange(
        self, public_key: t.Union[X25519PublicKey, bytes, str]
    ) -> bytes:
        """
        Takes in a public key from a communicating party & uses the
        instance's secret key to do an elliptic curve diffie-hellman
        exchange & returns the resulting secret shared bytes.
        """
        public_key = self._process_public_key(public_key)
        return self._Curve25519.exchange(
            self._secret_key, self._Curve25519.public_bytes(public_key)
        )

    @classmethod
    @comprehension()
    async def adh2_client(cls) -> Comprende:
        """
        Generates an ephemeral ``X25519`` secret key which is used to
        start a 2DH client key exchange. This key is yielded as public
        key bytes. Then the server's two public keys should be sent into
        this coroutine when they're received. When this coroutine
        reaches the raise statement, a primed ``sha3_512`` kdf object
        will be accessible from the ``aresult`` method of this generator.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import X25519

        async with X25519.protocols.adh2_client() as exchange:
            client_hello = await exchange()
            response = internet.post(client_hello)
            await exchange(response)

        shared_key_kdf = await exchange.aresult()
        """
        domain = Domains.DH2
        my_ephemeral_key = await cls().agenerate()
        peer_identity_key, peer_ephemeral_key = yield (
            my_ephemeral_key.public_bytes
        )
        shared_key_ad = await my_ephemeral_key.aexchange(peer_identity_key)
        shared_key_cd = await my_ephemeral_key.aexchange(peer_ephemeral_key)
        raise Comprende.ReturnValue(
            DomainKDF(
                domain,
                my_ephemeral_key.public_bytes,
                peer_identity_key,
                peer_ephemeral_key,
                key=shared_key_ad + shared_key_cd,
            )
        )

    @classmethod
    @comprehension()
    def dh2_client(cls) -> Comprende:
        """
        Generates an ephemeral ``X25519`` secret key which is used to
        start a 2DH client key exchange. This key is yielded as public
        key bytes. Then the server's two public keys should be sent into
        this coroutine when they're received. When this coroutine
        reaches the return statement, a primed ``sha3_512`` kdf object
        will be accessible from the ``result`` method of this generator.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import X25519

        with X25519.protocols.dh2_client() as exchange:
            client_hello = exchange()
            response = internet.post(client_hello)
            exchange(response)

        shared_key_kdf = exchange.result()
        """
        domain = Domains.DH2
        my_ephemeral_key = cls().generate()
        peer_identity_key, peer_ephemeral_key = yield (
            my_ephemeral_key.public_bytes
        )
        shared_key_ad = my_ephemeral_key.exchange(peer_identity_key)
        shared_key_cd = my_ephemeral_key.exchange(peer_ephemeral_key)
        return DomainKDF(
            domain,
            my_ephemeral_key.public_bytes,
            peer_identity_key,
            peer_ephemeral_key,
            key=shared_key_ad + shared_key_cd,
        )

    @comprehension()
    async def adh2_server(self, peer_ephemeral_key: bytes) -> Comprende:
        """
        Takes in the user's ``X25519`` secret key & a peer's public key
        bytes to enact a 2DH key exchange. This yields the user's two
        public keys as bytes, one from the secret key which was passed
        in as an argument, one which is ephemeral. When this coroutine
        reaches the raise statement, a primed ``sha3_512`` kdf object
        will be accessible from the ``aresult`` method of this generator.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import X25519

        # This key may be used again as an identity for this protocol
        ecdhe_key = await X25519().agenerate()

        client_ephemeral_key = internet.receive()

        async with ecdhe_key.adh2_server(client_ephemeral_key) as exchange:
            internet.send(await exchange.aexhaust())

        shared_key_kdf = await exchange.aresult()
        """
        domain = Domains.DH2
        my_identity_key = self
        my_ephemeral_key = await self.__class__().agenerate()
        yield my_identity_key.public_bytes, my_ephemeral_key.public_bytes
        shared_key_ad = await my_identity_key.aexchange(peer_ephemeral_key)
        shared_key_cd = await my_ephemeral_key.aexchange(peer_ephemeral_key)
        raise Comprende.ReturnValue(
            DomainKDF(
                domain,
                peer_ephemeral_key,
                my_identity_key.public_bytes,
                my_ephemeral_key.public_bytes,
                key=shared_key_ad + shared_key_cd,
            )
        )

    @comprehension()
    def dh2_server(self, peer_ephemeral_key: bytes) -> Comprende:
        """
        Takes in the user's ``X25519`` secret key & a peer's public key
        bytes to enact a 2DH key exchange. This yields the user's two
        public keys as bytes, one from the secret key which was passed
        in as an argument, one which is ephemeral. When this coroutine
        reaches the return statement, a primed ``sha3_512`` kdf object
        will be accessible from the ``result`` method of this generator.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import X25519

        # This key may be used again as an identity for this protocol
        ecdhe_key = X25519().generate()

        client_ephemeral_key = internet.receive()

        with ecdhe_key.dh2_server(client_ephemeral_key) as exchange:
            internet.send(exchange.exhaust())

        shared_key_kdf = exchange.result()
        """
        domain = Domains.DH2
        my_identity_key = self
        my_ephemeral_key = self.__class__().generate()
        yield my_identity_key.public_bytes, my_ephemeral_key.public_bytes
        shared_key_ad = my_identity_key.exchange(peer_ephemeral_key)
        shared_key_cd = my_ephemeral_key.exchange(peer_ephemeral_key)
        return DomainKDF(
            domain,
            peer_ephemeral_key,
            my_identity_key.public_bytes,
            my_ephemeral_key.public_bytes,
            key=shared_key_ad + shared_key_cd,
        )

    @comprehension()
    async def adh3_client(self) -> Comprende:
        """
        Takes in the user's ``X25519`` secret key to enact a 3DH key
        exchange with a peer. This yields the user's two public keys as
        bytes, one from the secret key which was passed in as an
        argument, one which is ephemeral. When this coroutine reaches
        the raise statement, a primed ``DomainKDF`` object will be
        accessible from the ``aresult`` method of this generator.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import X25519

        # This key may be used again as an identity for this protocol
        ecdhe_key = await X25519().agenerate()

        async with ecdhe_key.adh3_client() as exchange:
            client_hello = await exchange()
            response = internet.post(client_hello)
            await exchange(response)

        shared_key_kdf = await exchange.aresult()
        """
        domain = Domains.DH3
        my_identity_key = self
        my_ephemeral_key = await self.__class__().agenerate()
        peer_identity_key, peer_ephemeral_key = yield (
            my_identity_key.public_bytes,
            my_ephemeral_key.public_bytes,
        )
        shared_key_ad = await my_ephemeral_key.aexchange(peer_identity_key)
        shared_key_bc = await my_identity_key.aexchange(peer_ephemeral_key)
        shared_key_cd = await my_ephemeral_key.aexchange(peer_ephemeral_key)
        raise Comprende.ReturnValue(
            DomainKDF(
                domain,
                my_identity_key.public_bytes,
                my_ephemeral_key.public_bytes,
                peer_identity_key,
                peer_ephemeral_key,
                key=shared_key_ad + shared_key_bc + shared_key_cd,
            )
        )

    @comprehension()
    def dh3_client(self) -> Comprende:
        """
        Takes in the user's ``X25519`` secret key to enact a 3DH key
        exchange with a peer. This yields the user's two public keys as
        bytes, one from the secret key which was passed in as an
        argument, one which is ephemeral. When this coroutine reaches
        the return statement, a primed ``DomainKDF`` object will be
        accessible from the ``result`` method of this generator.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import X25519

        # This key may be used again as an identity for this protocol
        ecdhe_key = X25519().generate()

        with ecdhe_key.dh3_client() as exchange:
            client_hello = exchange()
            response = internet.post(client_hello)
            exchange(response)

        shared_key_kdf = exchange.result()
        """
        domain = Domains.DH3
        my_identity_key = self
        my_ephemeral_key = self.__class__().generate()
        peer_identity_key, peer_ephemeral_key = yield (
            my_identity_key.public_bytes,
            my_ephemeral_key.public_bytes,
        )
        shared_key_ad = my_ephemeral_key.exchange(peer_identity_key)
        shared_key_bc = my_identity_key.exchange(peer_ephemeral_key)
        shared_key_cd = my_ephemeral_key.exchange(peer_ephemeral_key)
        return DomainKDF(
            domain,
            my_identity_key.public_bytes,
            my_ephemeral_key.public_bytes,
            peer_identity_key,
            peer_ephemeral_key,
            key=shared_key_ad + shared_key_bc + shared_key_cd,
        )

    @comprehension()
    async def adh3_server(
        self, peer_identity_key: bytes, peer_ephemeral_key: bytes
    ) -> Comprende:
        """
        Takes in the user's ``X25519`` secret key & two of a peer's
        public keys bytes to enact a 3DH deniable key exchange. This
        yields the user's two public keys as bytes, one from the secret
        key which was passed in as an argument, one which is ephemeral.
        When this coroutine reaches the raise statement, a primed
        ``DomainKDF`` object will be accessible from the ``aresult``
        method of this generator.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import X25519

        # This key may be used again as an identity for this protocol
        ecdhe_key = await X25519().agenerate()

        identity_key, ephemeral_key = client_keys = internet.receive()

        server = ecdhe_key.adh3_server(identity_key, ephemeral_key)
        async with server as exchange:
            internet.send(await exchange.aexhaust())

        shared_key_kdf = await exchange.aresult()
        """
        domain = Domains.DH3
        my_identity_key = self
        my_ephemeral_key = await self.__class__().agenerate()
        yield my_identity_key.public_bytes, my_ephemeral_key.public_bytes
        shared_key_ad = await my_identity_key.aexchange(peer_ephemeral_key)
        shared_key_bc = await my_ephemeral_key.aexchange(peer_identity_key)
        shared_key_cd = await my_ephemeral_key.aexchange(peer_ephemeral_key)
        raise Comprende.ReturnValue(
            DomainKDF(
                domain,
                peer_identity_key,
                peer_ephemeral_key,
                my_identity_key.public_bytes,
                my_ephemeral_key.public_bytes,
                key=shared_key_ad + shared_key_bc + shared_key_cd,
            )
        )

    @comprehension()
    def dh3_server(
        self, peer_identity_key: bytes, peer_ephemeral_key: bytes
    ) -> Comprende:
        """
        Takes in the user's ``X25519`` secret key & two of a peer's
        public keys bytes to enact a 3DH deniable key exchange. This
        yields the user's two public keys as bytes, one from the secret
        key which was passed in as an argument, one which is ephemeral.
        When this coroutine reaches the raise statement, a primed
        ``DomainKDF`` object will be accessible from the ``result``
        method of this generator.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import X25519

        # This key may be used again as an identity for this protocol
        ecdhe_key = X25519().generate()

        identity_key, ephemeral_key = client_keys = internet.receive()

        server = ecdhe_key.dh3_server(identity_key, ephemeral_key)
        with server as exchange:
            internet.send(exchange.exhaust())

        shared_key_kdf = exchange.result()
        """
        domain = Domains.DH3
        my_identity_key = self
        my_ephemeral_key = self.__class__().generate()
        yield my_identity_key.public_bytes, my_ephemeral_key.public_bytes
        shared_key_ad = my_identity_key.exchange(peer_ephemeral_key)
        shared_key_bc = my_ephemeral_key.exchange(peer_identity_key)
        shared_key_cd = my_ephemeral_key.exchange(peer_ephemeral_key)
        return DomainKDF(
            domain,
            peer_identity_key,
            peer_ephemeral_key,
            my_identity_key.public_bytes,
            my_ephemeral_key.public_bytes,
            key=shared_key_ad + shared_key_bc + shared_key_cd,
        )


class PackageSigner:
    """
    Provides an intuitive API for users to sign their own packages.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    import json
    from getpass import getpass
    from aiootp import PackageSigner

    signer = PackageSigner(
        package=__package__,
        version=__version__,
        author=__author__,
        license=__license__,
    )
    signer.connect_to_secure_database(
        passphrase=getpass("database passphrase:\n"),
        salt=getpass("database salt:\n"),
        path=getpass("secure directory:\n"),
    )

    with open("MANIFEST.in", "r") as manifest:
        filename_sheet = manifest.read().split("\n")

    for line in filename_sheet:
        if not line.startswith("include"):
            continue
        filename = line.strip().split(" ")[-1]
        with open(filename, "rb") as source_file:
            signer.add_file(filename, source_file.read())

    signer.sign_package()
    signer.db.save_database()
    package_signature_summary = signer.summarize()
    """

    __slots__ = ("_db", "_scope", "files")

    _Hasher = sha384
    _Signer = Ed25519

    _CHECKSUM = CHECKSUM
    _CHECKSUMS = CHECKSUMS
    _CLASS = "PackageSigner"
    _PUBLIC_CREDENTIALS = PUBLIC_CREDENTIALS
    _SCOPE = SCOPE
    _SIGNATURE = SIGNATURE
    _SIGNING_KEY = SIGNING_KEY
    _VERSIONS = VERSIONS

    InvalidSignature = Ed25519._exceptions.InvalidSignature

    @classmethod
    def _database_template(cls) -> t.Dict[str, t.Union[str, t.JSONObject]]:
        """
        Returns the default instance package database values for
        initializing new databases.
        """
        return {
            cls._SIGNING_KEY: "",
            cls._VERSIONS: {},
            cls._PUBLIC_CREDENTIALS: {},
        }

    @classmethod
    def generate_signing_key(cls) -> _Signer:
        """
        Generates a new `Ed25519` secret signing key object.
        """
        return cls._Signer().generate()

    def __init__(
        self,
        package: str,
        version: str,
        date: t.Optional[int] = None,
        **scopes: t.JSONObject,
    ) -> "self":
        """
        Sets the instance's package scope attributes & default file
        checksums container.
        """
        self.files = PackageSignerFiles()
        self._scope = PackageSignerScope(
            package=package,
            version=version,
            date=date if date else day_clock.time(),
            **scopes,
        )

    def __repr__(self) -> str:
        """
        Displays the instance's declared scope.
        """
        cls = self.__class__.__qualname__
        return str(f"{cls}({self._scope})")

    @property
    def _public_credentials(self) -> t.JSONObject:
        """
        Returns public credentials from the instance's secure database.
        """
        return self.db[self._scope.package][self._PUBLIC_CREDENTIALS]

    @property
    def signing_key(self) -> _Signer:
        """
        Returns the package's secret signing key from the instance's
        encrypted database in an `Ed25519` object.
        """
        package = self._scope.package
        aad = Domains.PACKAGE_SIGNER + package.encode()
        encrypted_key = self.db[package][self._SIGNING_KEY]
        if not encrypted_key:
            raise PackageSignerIssue.signing_key_hasnt_been_set()
        key = self.db.read_token(encrypted_key, aad=aad)
        return self._Signer().import_secret_key(key)

    @property
    def db(self) -> "Database":
        """
        Returns the instance's database object, or alerts the user to
        connect to a secure database if it isn't yet set.
        """
        try:
            return self._db
        except AttributeError:
            raise PackageSignerIssue.must_connect_to_secure_database()

    @property
    def _checksums(self) -> t.Dict[str, str]:
        """
        Returns the instance's package filenames & their hexdigests in
        a JSON ready dictionary.
        """
        return {
            filename: hasher.hexdigest()
            for filename, hasher in sorted(self.files.items())
        }

    @property
    def _summary(self) -> t.JSONObject:
        """
        Collects the instance's package file checksums, names, public
        credentials, version scopes & the package's public signing key
        into a json-ready dictionary.
        """
        return {
            self._CHECKSUMS: {
                filename: hexdigest
                for filename, hexdigest in self._checksums.items()
            },
            self._PUBLIC_CREDENTIALS: {
                name: credential
                for name, credential in sorted(
                    self._public_credentials.items()
                )
            },
            self._SCOPE: {
                name: value for name, value in sorted(self._scope.items())
            },
            self._SIGNING_KEY: self.signing_key.public_bytes.hex(),
        }

    @property
    def _checksum(self) -> bytes:
        """
        Returns the digest of the current package summary.
        """
        return self._Hasher(json.dumps(self._summary).encode()).digest()

    @property
    def _signature(self) -> bytes:
        """
        Returns the stored package summary signature.
        """
        try:
            versions = self.db[self._scope.package][self._VERSIONS]
            return bytes.fromhex(versions[self._scope.version])
        except KeyError:
            raise PackageSignerIssue.package_hasnt_been_signed()

    def connect_to_secure_database(
        self,
        *secret_credentials: t.Iterable[bytes],
        username: bytes,
        passphrase: bytes,
        salt: bytes = b"",
        path: t.OptionalPathStr = None,
        **passcrypt_settings: t.PasscryptKWsNew,
    ) -> "self":
        """
        Opens an encrypted database connection using the Passcrypt
        passphrase-based key derivation function, a ``passphrase`` & any
        available additional credentials a user may have. If a database
        doesn't already exist, then a new one is created with default
        values.
        """
        from .databases import Database

        self._db = Database.generate_profile(
            self._CLASS.encode(),
            *secret_credentials,
            username=username,
            passphrase=passphrase,
            salt=salt,
            path=path,
            **passcrypt_settings,
        )
        try:
            self.db.query_tag(self._scope.package, cache=True)
        except LookupError:
            self.db[self._scope.package] = self._database_template()
        finally:
            return self

    def update_scope(self, **scopes) -> "self":
        """
        Updates the package scopes to qualify the package signature of
        the current package version within the instance.
        """
        self._scope.namespace.update(scopes)
        return self

    def update_public_credentials(
        self, **credentials: t.Dict[str, t.JSONSerializable]
    ) -> "self":
        """
        Updates the public credentials to be associated with the package
        signature & stores them in the instance's database cache. The
        database must be saved separately to save them to disk.
        """
        package = self._scope.package
        self.db[package][self._PUBLIC_CREDENTIALS].update(credentials)
        return self

    def update_signing_key(
        self, signing_key: t.Union[str, bytes, Ed25519PrivateKey, Ed25519]
    ) -> "self":
        """
        Updates the package's secret signing key as an encrypted token
        within the instance's database cache. The database must be saved
        separately to save the encrypted signing key to disk.
        """
        if signing_key.__class__ is not self._Signer:
            signing_key = self._Signer().import_secret_key(signing_key)
        package = self._scope.package
        aad = Domains.PACKAGE_SIGNER + package.encode()
        self.db[package][self._SIGNING_KEY] = self.db.make_token(
            signing_key.secret_bytes, aad=aad
        ).decode()
        return self

    def add_file(self, filename: str, file_data: bytes) -> "self":
        """
        Stores a ``filename`` & the hash object of the file's bytes type
        contents in the instance's `files` attribute mapping.
        """
        self.files[filename] = self._Hasher(file_data)
        return self

    def sign_package(self) -> "self":
        """
        Signs the package summary checksum & stores it in the instance's
        secure database cache. The database must be saved separately to
        save the signature to disk.
        """
        checksum = self._checksum
        self.db[self._scope.package][self._VERSIONS].update(
            {self._scope.version: self.signing_key.sign(checksum).hex()}
        )
        return self

    def summarize(self) -> t.JSONObject:
        """
        Assures the stored package checksum signature matches the
        current checksum of the package summary. If valid, the summary
        is returned along with the package checksum & its signature in
        a dictionary.
        """
        checksum = self._checksum
        signature = self._signature
        signing_key = self.signing_key
        try:
            signing_key.verify(signature, checksum)
        except self.InvalidSignature:
            raise PackageSignerIssue.out_of_sync_package_signature()
        return {
            self._CHECKSUM: checksum.hex(),
            **self._summary,
            self._SIGNATURE: signature.hex(),
        }


class PackageVerifier:
    """
    Provides an intuitive API for verifying package summaries produced
    by `PackageSigner` objects.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp import PackageVerifier

    verifier = PackageVerifier(public_signing_key, path="")
    verifier.verify_summary(package_signature_summary)
    """

    __slots__ = (
        "_path",
        "_checksum",
        "_signature",
        "_signing_key",
        "_summary_dictionary",
    )

    _Hasher = sha384
    _Signer = Ed25519

    _CHECKSUM = CHECKSUM
    _CHECKSUMS = CHECKSUMS
    _SIGNATURE = SIGNATURE
    _SIGNING_KEY = SIGNING_KEY

    InvalidSignature = Ed25519._exceptions.InvalidSignature

    def __init__(
        self,
        public_signing_key: t.Union[str, bytes, Ed25519PublicKey, Ed25519],
        *,
        path: t.OptionalPathStr = None,
        verify_files: bool = True,
    ) -> "self":
        """
        Receives the bytes type public signing key a user expects a
        package to be signed by, & stores it within the instance. The
        ``path`` keyword argument is the root location where all of the
        source files can be reached via the relative paths of the files
        declared in the summary. If instead the source files will not be
        checked for validity & only the validity of the signature will
        be assertained, the ``verify_files`` keyword can be set falsey.
        """
        if public_signing_key.__class__ is not self._Signer:
            key = self._Signer().import_public_key(public_signing_key)
        else:
            key = public_signing_key
        self._signing_key = key
        if verify_files:
            self._path = Path(path).absolute()
        elif path:
            raise Issue.unused_parameters("path", "not verifying files")

    @property
    def _summary_bytes(self) -> bytes:
        """
        Returns the UTF-8 encoded JSON package signature summary sans
        the package checksum & signature for hashing.
        """
        return json.dumps(self._summary_dictionary).encode()

    def _import_summary(
        self, summary: t.Dict[str, t.JSONSerializable]
    ) -> None:
        """
        Verifies the package summary checksum & stores its values within
        the instance.
        """
        summary = self._summary_dictionary = {**summary}
        self._checksum = bytes.fromhex(summary.pop(self._CHECKSUM))
        self._signature = bytes.fromhex(summary.pop(self._SIGNATURE))
        if self._Hasher(self._summary_bytes).digest() != self._checksum:
            raise Issue.invalid_value("package summary checksum")

    def _verify_file_checksums(self, summary: dict) -> None:
        """
        Verifies the files declared in the summary by loading them from
        the filesystem & calculating their digests for a match.
        """
        path = self._path
        files = summary[self._CHECKSUMS]
        for file_path, purported_hexdigest in files.items():
            purported_digest = bytes.fromhex(purported_hexdigest)
            with open(path / file_path, "rb") as source_file:
                digest = self._Hasher(source_file.read()).digest()
                if not bytes_are_equal(purported_digest, digest):
                    raise PackageSignerIssue.invalid_file_digest(file_path)

    def verify_summary(
        self,
        summary: t.Union[
            t.Dict[str, t.JSONSerializable], t.JSONDeserializable
        ],
    ) -> None:
        """
        Verifies the purported checksum of a package summary & the
        signature of the checksum.
        """
        if summary.__class__ is not dict:
            summary = json.loads(summary)
        purported_signing_key = bytes.fromhex(summary[self._SIGNING_KEY])
        if purported_signing_key != self._signing_key.public_bytes:
            raise Issue.invalid_value("summary's public signing key")
        self._import_summary(summary)
        if hasattr(self, "_path"):
            self._verify_file_checksums(summary)
        self._signing_key.verify(self._signature, self._checksum)


extras = dict(
    _KeyAADBundle=KeyAADBundle,
    DomainKDF=DomainKDF,
    Ed25519=Ed25519,
    PackageSigner=PackageSigner,
    PackageVerifier=PackageVerifier,
    Passcrypt=Passcrypt,
    X25519=X25519,
    __doc__=__doc__,
    __package__=__package__,
    abytes_keys=abytes_keys,
    agenerate_key=agenerate_key,
    amnemonic=amnemonic,
    bytes_keys=bytes_keys,
    generate_key=generate_key,
    mnemonic=mnemonic,
)


keygens = make_module("keygens", mapping=extras)



# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = []


__doc__ = (
    "A sub-package to implement functionalities to engage with users & "
    "security researchers."
)


from ._report_security_issue import report_security_issue




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *


async def test_detection_of_ciphertext_modification():
    aciphertext = Ciphertext(await cipher.abytes_encrypt(plaintext_bytes))

    ######
    ###### ciphertext container packs async ciphertexts correctly
    act = b"".join(aciphertext.values())
    assert all(aciphertext.values())
    assert act == b"".join(
        [aciphertext.shmac, aciphertext.salt, aciphertext.iv, aciphertext.ciphertext]
    )

    # async ciphertext doesn't obviously contain plaintext
    for chunk in gentools.data.root(plaintext_bytes, size=10):
        assert chunk not in act
    byte_leakage not in act

    # async decryption of correct data doesn't fail
    assert plaintext_bytes == await cipher.abytes_decrypt(act)

    # sync decryption of async ciphertext doesn't fail
    assert plaintext_bytes == cipher.bytes_decrypt(act)

    ######
    ###### async decryption of altered ciphertext fails
    aict = int.from_bytes(act, BYTE_ORDER)
    context = "Async ciphertext alteration not caught!"
    for abit in range(0, aict.bit_length(), 16):
        with ignore(StreamHMAC.InvalidSHMAC, if_else=violation(context)):
            altered_act = (aict ^ (1 << abit)).to_bytes(len(act), BYTE_ORDER)
            await cipher.abytes_decrypt(altered_act)

    # async decryption of ciphertext lengthened to invalid size fails
    context = "Invalid size lengthened sync ciphertext allowed"
    for extra_bytes in range(1, BLOCKSIZE):
        with ignore(ValueError, if_else=violation(context)):
            await cipher.abytes_decrypt(act + token_bytes(extra_bytes))

    # sync decryption of ciphertext shortened to invalid size fails
    context = "Invalid size shortened sync ciphertext allowed"
    for fewer_bytes in range(1, BLOCKSIZE):
        with ignore(ValueError, if_else=violation(context)):
            await cipher.abytes_decrypt(act[:-fewer_bytes])

    ######
    ###### test the ciphertext container class
    ciphertext = Ciphertext(cipher.bytes_encrypt(plaintext_bytes))

    # ciphertext container packs sync ciphertexts correctly
    ct = b"".join(ciphertext.values())
    assert all(ciphertext.values())
    assert ct == b"".join(
        [ciphertext.shmac, ciphertext.salt, ciphertext.iv, ciphertext.ciphertext]
    )

    # sync ciphertext doesn't obviously contain plaintext
    for chunk in gentools.data.root(plaintext_bytes, size=10):
        assert chunk not in ct
    byte_leakage not in ct

    # sync decryption of correct data doesn't fail
    assert plaintext_bytes == cipher.bytes_decrypt(ct)

    # async decryption of sync ciphertext doesn't fail
    assert plaintext_bytes == await cipher.abytes_decrypt(ct)

    ######
    ###### sync decryption of altered ciphertext fails
    ict = int.from_bytes(ct, BYTE_ORDER)
    context = "Sync ciphertext alteration not caught!"
    for bit in range(0, ict.bit_length(), 16):
        with ignore(StreamHMAC.InvalidSHMAC, if_else=violation(context)):
            altered_ct = (ict ^ (1 << bit)).to_bytes(len(ct), BYTE_ORDER)
            cipher.bytes_decrypt(altered_ct)

    # sync decryption of ciphertext lengthened to invalid size fails
    context = "Invalid size lengthened sync ciphertext allowed"
    for extra_bytes in range(1, BLOCKSIZE):
        with ignore(ValueError, if_else=violation(context)):
            cipher.bytes_decrypt(ct + token_bytes(extra_bytes))

    # sync decryption of ciphertext shortened to invalid size fails
    context = "Invalid size shortened sync ciphertext allowed"
    for fewer_bytes in range(1, BLOCKSIZE):
        with ignore(ValueError, if_else=violation(context)):
            cipher.bytes_decrypt(ct[:-fewer_bytes])


async def aciphertext_stream():
    key_bundle = await KeyAADBundle(key, aad=aad).async_mode()
    shmac = StreamHMAC(key_bundle)._for_encryption()
    datastream = aplaintext_stream(plaintext_bytes)
    cipherstream = abytes_encipher(datastream, shmac=shmac)

    first_ciphertext_block = await cipherstream.asend(None)
    yield key_bundle.salt, key_bundle.iv
    yield (
        await shmac.anext_block_id(first_ciphertext_block),
        first_ciphertext_block,
    )
    async for ciphertext_block in cipherstream:
        yield (
            await shmac.anext_block_id(ciphertext_block),
            ciphertext_block,
        )


def ciphertext_stream():
    key_bundle = KeyAADBundle(key, aad=aad).sync_mode()
    shmac = enc_hmac = StreamHMAC(key_bundle)._for_encryption()
    datastream = plaintext_stream(plaintext_bytes)
    cipherstream = bytes_encipher(datastream, shmac=shmac)

    first_ciphertext_block = cipherstream.send(None)
    yield key_bundle.salt, key_bundle.iv
    yield (
        shmac.next_block_id(first_ciphertext_block),
        first_ciphertext_block,
    )
    for ciphertext_block in cipherstream:
        yield (
            shmac.next_block_id(ciphertext_block),
            ciphertext_block,
        )


async def test_async_block_ids_during_deciphering():
    cipherstream = aciphertext_stream()
    salt, iv = await cipherstream.asend(None)
    key_bundle = await KeyAADBundle(key, salt=salt, aad=aad, iv=iv).async_mode()
    shmac = StreamHMAC(key_bundle)._for_decryption()

    ciphertext = []
    deciphering = abytes_decipher(aunpack(ciphertext), shmac=shmac)

    padded_plaintext = b""
    async for block_id, ciphertext_block in cipherstream:
        await shmac.atest_next_block_id(block_id, ciphertext_block)
        ciphertext.append(ciphertext_block)
        padded_plaintext += await deciphering.asend(None)

        # altering the block_id fails
        context = "Block id was modified without notice!"
        with ignore(StreamHMAC.InvalidBlockID, if_else=violation(context)):
            fake_block_id = await axi_mix(block_id + b"\x01", size=BLOCK_ID_BYTES)
            await shmac.atest_next_block_id(fake_block_id, ciphertext_block)

        # a block_id that is too short fails
        context = "An insufficient size block id was allowed!"
        with ignore(PermissionError, if_else=violation(context)):
            truncated_block_id = block_id[:MIN_BLOCK_ID_BYTES - 1]
            await shmac.atest_next_block_id(truncated_block_id, ciphertext_block)

        # alterting the ciphertext_block fails
        context = "Block was modified without notice!"
        with ignore(StreamHMAC.InvalidBlockID, if_else=violation(context)):
            fake_block = await axi_mix(ciphertext_block + b"\x01", size=BLOCKSIZE)
            await shmac.atest_next_block_id(block_id, fake_block)

    assert plaintext_bytes == await Padding.adepad_plaintext(
        padded_plaintext
    )

    context = "MAC object accessible after finalization!"
    tag = await shmac.afinalize()
    await shmac.atest_shmac(tag)
    with ignore(PermissionError, if_else=violation(context)):
        shmac._mac.digest()


def test_sync_block_ids_during_deciphering():
    stream = ciphertext_stream()
    salt, iv = stream.send(None)
    key_bundle = KeyAADBundle(key, salt=salt, aad=aad, iv=iv).sync_mode()
    shmac = StreamHMAC(key_bundle)._for_decryption()

    ciphertext = []
    deciphering = bytes_decipher(unpack(ciphertext), shmac=shmac)

    padded_plaintext = b""
    for block_id, ciphertext_block in stream:
        shmac.test_next_block_id(block_id, ciphertext_block)
        ciphertext.append(ciphertext_block)
        padded_plaintext += deciphering.send(None)

        # altering the block_id fails
        context = "Block id was modified without notice!"
        with ignore(StreamHMAC.InvalidBlockID, if_else=violation(context)):
            fake_block_id = xi_mix(block_id + b"\x01", size=BLOCK_ID_BYTES)
            shmac.test_next_block_id(fake_block_id, ciphertext_block)

        # a block_id that is too short fails
        context = "An insufficient size block id was allowed!"
        with ignore(PermissionError, if_else=violation(context)):
            truncated_block_id = block_id[:MIN_BLOCK_ID_BYTES - 1]
            shmac.test_next_block_id(truncated_block_id, ciphertext_block)

        # alterting the ciphertext_block fails
        context = "Block was modified without notice!"
        with ignore(StreamHMAC.InvalidBlockID, if_else=violation(context)):
            fake_block = xi_mix(ciphertext_block + b"\x01", size=BLOCKSIZE)
            shmac.test_next_block_id(block_id, fake_block)

    assert plaintext_bytes == Padding.depad_plaintext(padded_plaintext)

    context = "MAC object accessible after finalization!"
    tag = shmac.finalize()
    shmac.test_shmac(tag)
    with ignore(PermissionError, if_else=violation(context)):
        shmac._mac.digest()


def test_sync_cipher_decipher_streams():
    # buffering protocol works for varying sizes
    for i in range(0, 512 - INNER_HEADER_BYTES - 1, unique_range(1, 64)):
        stream_enc = CipherStream(key)
        pt_enc = i * b"a"
        stream_enc.buffer(pt_enc)

        stream_dec = DecipherStream(
            key, salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
        )
        for var in ("salt", "iv", "aad", "PACKETSIZE"):
            assert getattr(stream_enc, var) == getattr(stream_dec, var)
        pt_dec = b""
        join = b"".join
        for id_ct in stream_enc:
            stream_dec.buffer(join(id_ct))
            for pt in stream_dec:
                pt_dec += pt
        for id_ct in stream_enc.finalize():
            stream_dec.buffer(join(id_ct))
        for pt in stream_dec.finalize():
            pt_dec += pt
        assert pt_dec == pt_enc


async def test_async_cipher_decipher_streams():
    # buffering protocol works for varying sizes
    for i in range(0, 512 - INNER_HEADER_BYTES - 1, unique_range(1, 64)):
        stream_enc = await AsyncCipherStream(key, salt=salt, aad=aad)
        pt_enc = i * b"a"
        await stream_enc.abuffer(pt_enc)

        stream_dec = await AsyncDecipherStream(
            key, salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
        )
        for var in ("salt", "iv", "aad", "PACKETSIZE"):
            assert getattr(stream_enc, var) == getattr(stream_dec, var)
        pt_dec = b""
        join = b"".join
        async for id_ct in stream_enc:
            await stream_dec.abuffer(join(id_ct))
            async for pt in stream_dec:
                pt_dec += pt
        async for id_ct in stream_enc.afinalize():
            await stream_dec.abuffer(join(id_ct))
        async for pt in stream_dec.afinalize():
            pt_dec += pt
        assert pt_dec == pt_enc


async def test_async_encipher_sync_decipher_interop():
    # inter-op between async encryption & sync decryption doesn't fail
    stream_enc = await AsyncCipherStream(key, salt=salt, aad=aad)
    pt_enc = plaintext_bytes
    await stream_enc.abuffer(pt_enc)

    stream_dec = DecipherStream(
        key, salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
    )
    for var in ("salt", "iv", "aad", "PACKETSIZE"):
        assert getattr(stream_enc, var) == getattr(stream_dec, var)
    pt_dec = b""
    join = b"".join

    async for block_id, block in stream_enc:
        # the cipher can be resumed after the failure state of an
        # InvalidBlockID exception is recovered & the correct data
        # is supplied to the buffer
        context = "An altered block_id was not detected"
        fake_block_id = xi_mix(block_id + b"\x01", size=len(block_id))
        with ignore(InvalidBlockID, if_else=violation(context)) as relay:
            stream_dec.buffer(fake_block_id + block)
        assert fake_block_id == relay.error.failure_state.block_id
        assert block == relay.error.failure_state.block
        assert 0 == len(relay.error.failure_state.buffer())

        context = "An altered block was not detected"
        fake_block = xi_mix(block + b"\x01", size=len(block))
        with ignore(InvalidBlockID, if_else=violation(context)) as relay:
            stream_dec.buffer(block_id + fake_block)
        assert block_id == relay.error.failure_state.block_id
        assert fake_block == relay.error.failure_state.block
        assert 0 == len(relay.error.failure_state.buffer())

        # correct data is processed without failure
        stream_dec.buffer(block_id + block)
        for pt in stream_dec:
            pt_dec += pt
    async for id_ct in stream_enc.afinalize():
        stream_dec.buffer(join(id_ct))
    for pt in stream_dec.finalize():
        pt_dec += pt
    assert pt_dec == pt_enc


async def test_sync_encipher_async_decipher_interop():
    # inter-op between sync encryption & async decryption doesn't fail
    stream_enc = CipherStream(key)
    pt_enc = plaintext_bytes
    stream_enc.buffer(pt_enc)

    stream_dec = await AsyncDecipherStream(
        key, salt=stream_enc.salt, aad=stream_enc.aad, iv=stream_enc.iv
    )
    for var in ("salt", "iv", "aad", "PACKETSIZE"):
        assert getattr(stream_enc, var) == getattr(stream_dec, var)
    pt_dec = b""
    join = b"".join

    for block_id, block in stream_enc:
        # the cipher can be resumed after the failure state of an
        # InvalidBlockID exception is recovered & the correct data
        # is supplied to the buffer
        context = "An altered block_id was not detected"
        fake_block_id = xi_mix(block_id + b"\x01", size=len(block_id))
        async with aignore(InvalidBlockID, if_else=aviolation(context)) as relay:
            await stream_dec.abuffer(fake_block_id + block)
        assert fake_block_id == relay.error.failure_state.block_id
        assert block == relay.error.failure_state.block
        assert 0 == len(relay.error.failure_state.buffer())

        context = "An altered block was not detected"
        fake_block = xi_mix(block + b"\x01", size=len(block))
        async with aignore(InvalidBlockID, if_else=aviolation(context)) as relay:
            await stream_dec.abuffer(block_id + fake_block)
        assert block_id == relay.error.failure_state.block_id
        assert fake_block == relay.error.failure_state.block
        assert 0 == len(relay.error.failure_state.buffer())

        # correct data is processed without failure
        await stream_dec.abuffer(block_id + block)
        async for pt in stream_dec:
            pt_dec += pt
    for id_ct in stream_enc.finalize():
        await stream_dec.abuffer(join(id_ct))
    async for pt in stream_dec.afinalize():
        pt_dec += pt
    assert pt_dec == pt_enc


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = ["_containers", "_exceptions", "_typing"]


__doc__ = (
    "Coordinates some of the library's UI/UX by inserting higher-level "
    "functionality from dependant modules into lower-level modules in t"
    "his package."
)


from . import *
from . import _typing, _containers, _exceptions
from ._typing import Typing
from ._debuggers import gen_timer, agen_timer
from .asynchs import Processes, Threads, AsyncInit
from .commons import Namespace, OpenNamespace, Slots
from .commons import make_module
from .gentools import Comprende, comprehension
from .generics import Hasher, Domains, BytesIO, Padding
from .randoms import SequenceID
from .ciphers import StreamHMAC, SyntheticIV, Chunky2048
from .ciphers import generate_salt, agenerate_salt
from .keygens import KeyAADBundle
from .keygens import Ed25519, X25519, DomainKDF, Passcrypt
from .keygens import PackageSigner, PackageVerifier
from .databases import Database, AsyncDatabase


@comprehension()
async def adebugger(
    self: Comprende, *args, **kwargs
) -> Typing.AsyncGenerator:
    """
    Allows users to benchmark & read inspection details of running async
    generators inline as a chainable method.
    """
    args = args if args else self._args
    kwargs = {**self._kwargs, **kwargs}
    async for result in agen_timer(self._func)(*args, **kwargs):
        yield result


@comprehension()
def debugger(self: Comprende, *args, **kwargs) -> Typing.Generator:
    """
    Allows users to benchmark & read inspection details of running sync
    generators inline as a chainable method.
    """
    args = args if args else self._args
    kwargs = {**self._kwargs, **kwargs}
    yield from gen_timer(self._func)(*args, **kwargs)


def insert_debuggers() -> None:
    """
    Copies the addons over into the ``Comprende`` class.
    """
    addons = (debugger, adebugger)
    for addon in addons:
        setattr(Comprende, addon.__name__, addon)
        Comprende.lazy_generators.add(addon.__name__)


def insert_types() -> None:
    """
    Gives the package's type-hinting helper class access to the higher
    level classes.
    """
    Typing.AsyncCipherStream = AsyncCipherStream
    Typing.AsyncDecipherStream = AsyncDecipherStream
    Typing.AsyncInit = AsyncInit
    Typing.CipherStream = CipherStream
    Typing.DecipherStream = DecipherStream
    Typing.AsyncDatabase = AsyncDatabase
    Typing.BytesIO = BytesIO
    Typing.Chunky2048 = Chunky2048
    Typing.Comprende = Comprende
    Typing.Database = Database
    Typing.DomainKDF = DomainKDF
    Typing.Domains = Domains
    Typing.Ed25519 = Ed25519
    Typing.GUID = GUID
    Typing.Hasher = Hasher
    Typing.KeyAADBundle = KeyAADBundle
    Typing.Namespace = Namespace
    Typing.OpenNamespace = OpenNamespace
    Typing.PackageSigner = PackageSigner
    Typing.PackageVerifier = PackageVerifier
    Typing.Padding = Padding
    Typing.Passcrypt = Passcrypt
    Typing.PasscryptInstance = Passcrypt().__class__
    Typing.Processes = Processes
    Typing.SequenceID = SequenceID
    Typing.Slots = Slots
    Typing.StreamHMAC = StreamHMAC
    Typing.SyntheticIV = SyntheticIV
    Typing.Threads = Threads
    Typing.X25519 = X25519


def overwrite_containers_module() -> None:
    """
    Overwrites the package's `_containers.py` variable which will then
    be accessible to the user.
    """
    global _containers

    mapping = OpenNamespace(
        **{
            name: getattr(_containers, name) for name in _containers.__all__
        },
        __all__=_containers.__all__,
        __doc__=_containers.__doc__,
        __package__=_containers.__package__,
    )
    _containers = make_module("_containers", mapping=mapping)


def overwrite_exceptions_module() -> None:
    """
    Overwrites the package's `_exceptions.py` variable which will then
    be accessible to the user.
    """
    global _exceptions

    mapping = OpenNamespace(
        **{
            name: getattr(_exceptions, name) for name in _exceptions.__all__
        },
        __all__=_exceptions.__all__,
        __doc__=_exceptions.__doc__,
        __package__=_exceptions.__package__,
    )
    _exceptions = make_module("_exceptions", mapping=mapping)


def overwrite_typing_module() -> None:
    """
    Overwrites the package's `_typing.py` variable which will then be
    accessible to the user.
    """
    global _typing

    mapping = OpenNamespace(
        Typing=_typing.Typing,
        __all__=_typing.__all__,
        __doc__=_typing.__doc__,
        __package__=_typing.__package__,
    )
    _typing = make_module("_typing", mapping=mapping)


insert_debuggers()
insert_types()
overwrite_containers_module()
overwrite_exceptions_module()
overwrite_typing_module()




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *


class TestDatabaseCacheSystem:
    async def test_async_clear_cache_clears_metatags_when_instructed(
        self, async_database
    ):
        """
        Setup: 1) a metatag exists AND both the metatag & its parent
        have tags in their caches & saved on the filesystem.

        Test: 2) clearing a database's cache clears its metatags' caches
        when instructed.

        Test: 3) clearing a database's cache doesn't clear its metatags'
        caches when instructed not to.
        """
        # 1
        child = await async_database.ametatag(metatag)
        child[tag] = plaintext_bytes
        async_database[tag] = test_data
        test_data_copy = test_data.copy()
        await async_database.asave_database()

        assert plaintext_bytes
        assert plaintext_bytes.__class__ is bytes
        assert child[tag] == plaintext_bytes
        assert test_data
        assert test_data.__class__ is dict
        assert async_database[tag] == test_data

        # 2
        await async_database.aclear_cache(metatags=True)
        assert plaintext_bytes
        assert plaintext_bytes.__class__ is bytes
        assert child[tag] != plaintext_bytes
        assert child[tag] == None
        assert test_data_copy == test_data
        assert test_data.__class__ is dict
        assert async_database[tag] != test_data
        assert async_database[tag] == None

        # 3
        await async_database.aquery_tag(tag, cache=True)
        assert async_database[tag] == test_data

        await child.aquery_tag(tag, cache=True)
        assert child[tag] == plaintext_bytes

        await async_database.aclear_cache(metatags=False)
        assert plaintext_bytes
        assert plaintext_bytes.__class__ is bytes
        assert child[tag] == plaintext_bytes
        assert test_data_copy == test_data
        assert test_data.__class__ is dict
        assert async_database[tag] != test_data
        assert async_database[tag] == None

    def test_clear_cache_clears_metatags_when_instructed(self, database):
        """
        Setup: 1) a metatag exists AND both the metatag & its parent
        have tags in their caches & saved on the filesystem.

        Test: 2) clearing a database's cache clears its metatags' caches
        when instructed.

        Test: 3) clearing a database's cache doesn't clear its metatags'
        caches when instructed not to.
        """
        # 1
        child = database.metatag(metatag)
        child[tag] = plaintext_bytes
        database[tag] = test_data
        test_data_copy = test_data.copy()
        database.save_database()

        assert plaintext_bytes
        assert plaintext_bytes.__class__ is bytes
        assert child[tag] == plaintext_bytes
        assert test_data
        assert test_data.__class__ is dict
        assert database[tag] == test_data

        # 2
        database.clear_cache(metatags=True)
        assert plaintext_bytes
        assert plaintext_bytes.__class__ is bytes
        assert child[tag] != plaintext_bytes
        assert child[tag] == None
        assert test_data_copy == test_data
        assert test_data.__class__ is dict
        assert database[tag] != test_data
        assert database[tag] == None

        # 3
        database.query_tag(tag, cache=True)
        assert database[tag] == test_data

        child.query_tag(tag, cache=True)
        assert child[tag] == plaintext_bytes

        database.clear_cache(metatags=False)
        assert plaintext_bytes
        assert plaintext_bytes.__class__ is bytes
        assert child[tag] == plaintext_bytes
        assert test_data_copy == test_data
        assert test_data.__class__ is dict
        assert database[tag] != test_data
        assert database[tag] == None


def test_Database_instance(database):
    db = Database(key=key, preload=True)

    # basic database functionalities work the same across reloads
    assert db._Database__root_kdf.sha3_256() == database._Database__root_kdf.sha3_256()
    assert db._Database__kdf.sha3_256() == database._Database__kdf.sha3_256()
    assert db._Database__root_kdf.sha3_256() != database._Database__kdf.sha3_256()
    assert db._Database__root_salt == database._Database__root_salt
    assert db._root_filename == database._root_filename
    assert db._root_salt_filename == database._root_salt_filename
    assert db.make_hmac(atag.encode()) == database.make_hmac(atag.encode())
    assert db.filename(atag) == database.filename(atag)
    assert db._metatag_key(atag) == database._metatag_key(atag)


async def test_AsyncDatabase_instance(database):
    db = await AsyncDatabase(key=key, preload=True)

    # basic async database functionalities work the same across reloads
    assert db._AsyncDatabase__root_kdf.sha3_256() == database._Database__root_kdf.sha3_256()
    assert db._AsyncDatabase__kdf.sha3_256() == database._Database__kdf.sha3_256()
    assert db._AsyncDatabase__root_kdf.sha3_256() != database._Database__kdf.sha3_256()
    assert db._AsyncDatabase__root_salt == database._Database__root_salt
    assert db._root_filename == database._root_filename
    assert db._root_salt_filename == database._root_salt_filename
    assert await db.amake_hmac(tag.encode()) == database.make_hmac(tag.encode())
    assert await db.afilename(tag) == database.filename(tag)
    assert await db._ametatag_key(tag) == database._metatag_key(tag)


def test_Database_user_kdf(database):
    db = Database(key=key, preload=False)

    context = "empty kdf updates were allowed!"
    with ignore(ValueError, if_else=violation(context)):
        db.kdf.update()

    assert db.kdf.sha3_256() == database.kdf.sha3_256()
    assert db.kdf.sha3_512() == database.kdf.sha3_512()
    assert db.kdf.shake_128(32) == database.kdf.shake_128(32)
    assert db.kdf.shake_256(32) == database.kdf.shake_256(32)


async def test_AsyncDatabase_user_kdf(async_database):
    db = await AsyncDatabase(key=key, preload=False)

    context = "empty kdf updates were allowed!"
    with ignore(ValueError, if_else=violation(context)):
        await db.kdf.aupdate()

    assert await db.kdf.asha3_256() == await async_database.kdf.asha3_256()
    assert await db.kdf.asha3_512() == await async_database.kdf.asha3_512()
    assert await db.kdf.ashake_128(32) == await async_database.kdf.ashake_128(32)
    assert await db.kdf.ashake_256(32) == await async_database.kdf.ashake_256(32)


def test_database_ciphers(database):
    # database ciphertexts are unique
    db = database
    filename = db.filename(tag)
    encrypted_data = db.json_encrypt(test_data, filename=filename, aad=DEFAULT_AAD)
    db[tag] = test_data
    db.save_database()
    encrypted_file = db._query_ciphertext(filename)
    assert encrypted_file != encrypted_data
    assert encrypted_file[SHMAC_SLICE] != encrypted_data[SHMAC_SLICE]
    assert encrypted_file[SALT_SLICE] != encrypted_data[SALT_SLICE]
    assert encrypted_file[IV_SLICE] != encrypted_data[IV_SLICE]
    assert encrypted_file[CIPHERTEXT_SLICE] != encrypted_data[CIPHERTEXT_SLICE]

    # database ciphers recover json data correctly
    assert test_data == db.json_decrypt(encrypted_file, filename=filename, aad=DEFAULT_AAD)
    assert test_data == db.json_decrypt(encrypted_data, filename=filename, aad=DEFAULT_AAD)
    assert type(test_data) is dict

    # database ciphers recover bytes data correctly
    encrypted_binary_data = db.bytes_encrypt(plaintext_bytes, aad=b"test")
    decrypted_binary_data = db.bytes_decrypt(
        encrypted_binary_data, aad=b"test", ttl=30
    )
    assert decrypted_binary_data == plaintext_bytes
    assert type(encrypted_binary_data) is bytes
    assert type(decrypted_binary_data) is bytes

    # database ciphers recover token data correctly
    encrypted_token_data = db.make_token(plaintext_bytes, aad=b"test")
    decrypted_token_data = db.read_token(
        encrypted_token_data, aad=b"test", ttl=3600
    )
    assert decrypted_token_data == plaintext_bytes
    assert type(encrypted_token_data) is bytes
    assert type(decrypted_token_data) is bytes
    assert (
        set(Tables.URL_SAFE.encode()).union(encrypted_token_data + b"%")
        == set(Tables.URL_SAFE.encode()).union(b"%")
    )


async def test_async_database_ciphers(async_database):
    # async database ciphertexts are unique
    db = async_database
    filename = await db.afilename(tag)
    encrypted_data = await db.ajson_encrypt(test_data, filename=filename, aad=DEFAULT_AAD)
    db[tag] = test_data
    await db.asave_database()
    encrypted_file = await db._aquery_ciphertext(filename)
    assert encrypted_file != encrypted_data
    assert encrypted_file[SHMAC_SLICE] != encrypted_data[SHMAC_SLICE]
    assert encrypted_file[SALT_SLICE] != encrypted_data[SALT_SLICE]
    assert encrypted_file[IV_SLICE] != encrypted_data[IV_SLICE]

    # async database ciphers recover json data correctly
    assert test_data == await db.ajson_decrypt(encrypted_file, filename=filename, aad=DEFAULT_AAD)
    assert test_data == await db.ajson_decrypt(encrypted_data, filename=filename, aad=DEFAULT_AAD)

    # async database ciphers recover bytes data correctly
    encrypted_binary_data = await db.abytes_encrypt(
        plaintext_bytes, aad=b"test"
    )
    decrypted_binary_data = await db.abytes_decrypt(
        encrypted_binary_data, aad=b"test", ttl=30
    )
    assert decrypted_binary_data == plaintext_bytes
    assert type(encrypted_binary_data) is bytes

    # async database ciphers recover token data correctly
    encrypted_token_data = await db.amake_token(plaintext_bytes, aad=b"test")
    decrypted_token_data = await db.aread_token(
        encrypted_token_data, aad=b"test", ttl=3600
    )
    assert decrypted_token_data == plaintext_bytes
    assert type(encrypted_token_data) is bytes
    assert (
        set(Tables.URL_SAFE.encode()).union(encrypted_token_data + b"%")
        == set(Tables.URL_SAFE.encode()).union(b"%")
    )


async def test_async_tags_metatags():
    async_database = await AsyncDatabase(key * 2, preload=True)
    achild = await async_database.ametatag(ametatag, preload=True)

    # async databases retrieve their stored data uncorrupted
    await async_database.aset_tag("bytes", plaintext_bytes, cache=False)
    assert await async_database.aquery_tag("bytes") == plaintext_bytes

    # metatag references stored as parent attributes are identical to
    # the references returned by their (a)metatag methods
    achild[atag] = atest_data
    assert achild is async_database.aclients
    assert achild[atag] is async_database.aclients[atag]

    # exiting an async database's async context manager saves their
    # childrens' files to disk
    afilename = await achild.afilename(atag)
    assert not (achild.path / afilename).exists()
    async with async_database:
        pass
    assert (achild.path / afilename).exists()

    # metatags of equivalent but not identical database instances
    # contain equivalent but not identical stored data
    adb = await AsyncDatabase(key * 2, preload=True)
    assert async_database.aclients[atag] == adb.aclients[atag]
    assert async_database.aclients[atag] is not adb.aclients[atag]

    # all databases create children from the keys returned by their
    # _ametatag_key methods
    ametatag_key = await async_database._ametatag_key(ametatag)
    achild = await AsyncDatabase(ametatag_key, metatag=True, preload=True)
    assert achild[atag]
    assert achild[atag].__class__ is dict
    assert achild[atag] == async_database.aclients[atag]
    assert achild[atag] is not async_database.aclients[atag]

    await async_database.adelete_database()


def test_sync_tags_metatags():
    database = Database(key * 2, preload=True)
    child = database.metatag(metatag, preload=True)

    # databases retrieve their stored data uncorrupted
    database.set_tag("bytes", plaintext_bytes, cache=False)
    assert database.query_tag("bytes") == plaintext_bytes

    # metatag references stored as parent attributes are identical to
    # the references returned by their (a)metatag methods
    child[tag] = test_data
    assert child is database.clients
    assert child[tag] is database.clients[tag]

    # exiting database's context manager saves their childrens' files to
    # disk
    filename = child.filename(tag)
    assert not (child.path / filename).exists()
    with database:
        pass
    assert (child.path / filename).exists()

    # metatags of equivalent but not identical database instances
    # contain equivalent but not identical stored data
    db = Database(key * 2, preload=True)
    assert database.clients[tag] == db.clients[tag]
    assert database.clients[tag] is not db.clients[tag]

    # all databases create children from the keys returned by their
    # _metatag_key methods
    metatag_key = database._metatag_key(metatag)
    child = Database(metatag_key, metatag=True, preload=True)
    assert child[tag]
    assert child[tag].__class__ is dict
    assert child[tag] == database.clients[tag]
    assert child[tag] is not database.clients[tag]

    database.delete_database()


async def test_async_user_profiles(async_database):
    adb = async_database
    user = await adb.agenerate_profile(**PROFILE_AND_SETTINGS)

    async with user:
        user[atag] = atest_data

    # equivalent async profiles contain their own copies of stored data
    user_copy = await adb.agenerate_profile(**PROFILE_AND_SETTINGS, preload=True)
    assert user[atag] == user_copy[atag]
    assert user.tags == user_copy.tags
    assert user[atag]
    assert user[atag] == atest_data
    assert user[atag].__class__ is dict
    assert user[atag] == user_copy[atag]
    assert user[atag] is not user_copy[atag]

    # async profiles are automatically saved to disk when initialized
    assert user_copy._root_path.is_file()
    assert user_copy._profile_tokens._salt_path.is_file()
    assert (user_copy.path / await user_copy.afilename(atag)).is_file()

    # async & sync profile contructors are equivalent
    sync_user = Database.generate_profile(**PROFILE_AND_SETTINGS, preload=True)
    assert sync_user[atag] == user[atag]
    assert sync_user.tags == user.tags
    assert sync_user[atag]
    assert sync_user[atag] == atest_data
    assert sync_user[atag].__class__ is dict
    assert sync_user[atag] is not user[atag]

    # deleting an async profile removes its files from the filesystem
    await user.adelete_database()
    assert not user_copy._root_path.is_file()
    assert not user_copy._profile_tokens._salt_path.is_file()
    assert not (user_copy.path / await user_copy.afilename(atag)).is_file()


async def test_user_profiles(database):
    db = database
    user = db.generate_profile(**PROFILE_AND_SETTINGS)

    with user:
        user[tag] = test_data

    # equivalent profiles contain their own copies of stored data
    user_copy = db.generate_profile(**PROFILE_AND_SETTINGS, preload=True)
    assert user[tag] == user_copy[tag]
    assert user.tags == user_copy.tags
    assert user[tag]
    assert user[tag] == atest_data
    assert user[tag].__class__ is dict
    assert user[tag] == user_copy[tag]
    assert user[tag] is not user_copy[tag]

    # profiles are automatically saved to disk when initialized
    assert user_copy._root_path.is_file()
    assert user_copy._profile_tokens._salt_path.is_file()
    assert (user_copy.path / user_copy.filename(tag)).is_file()

    # async & sync profile contructors are equivalent
    async_user = await AsyncDatabase.agenerate_profile(**PROFILE_AND_SETTINGS, preload=True)
    assert async_user[tag] == user[tag]
    assert async_user.tags == user.tags
    assert async_user[tag]
    assert async_user[tag] == test_data
    assert async_user[tag].__class__ is dict
    assert async_user[tag] is not user[tag]

    # deleting a profile removes its files from the filesystem
    user.delete_database()
    assert not user_copy._root_path.is_file()
    assert not user_copy._profile_tokens._salt_path.is_file()
    assert not (user_copy.path / user_copy.filename(tag)).is_file()


class TestHMACMethods:
    def test_sync_hmac_methods_are_sound(self, database):
        """
        Sync HMAC methods provide soundness of data validation.
        """
        inputs = token_bytes(32)
        tag = database.make_hmac(inputs)

        # sync validation doesn't fail
        database.test_hmac(tag, inputs)

        # sync hmac tags fail when inputs type is altered
        context = "Data type alteration not caught!"
        with ignore(TypeError, if_else=violation(context)):
            database.test_hmac(tag, str(inputs))

        # sync hmac tags fail when inputs are altered
        iinputs = int.from_bytes(inputs, BYTE_ORDER)
        context = "Data value alteration not caught!"
        for bit in range(iinputs.bit_length()):
            with ignore(database.InvalidHMAC, if_else=violation(context)):
                database.test_hmac(tag, (iinputs ^ (1 << bit)).to_bytes(32, BYTE_ORDER))

        # sync hmac tags fail when they are altered
        itag = int.from_bytes(tag, BYTE_ORDER)
        context = "Tag alteration not caught!"
        for bit in range(itag.bit_length()):
            with ignore(database.InvalidHMAC, if_else=violation(context)):
                altered_tag = (itag ^ (1 << bit)).to_bytes(32, BYTE_ORDER)
                database.test_hmac(altered_tag, inputs)

    async def test_async_hmac_methods_are_sound(self, async_database):
        """
        Async HMAC methods provide soundness of data validation.
        """
        ainputs = token_bytes(32)
        atag = await async_database.amake_hmac(ainputs)

        # async validation doesn't fail
        await async_database.atest_hmac(atag, ainputs)

        # async hmac tags fail when inputs type is altered
        context = "Async data type alteration not caught!"
        async with aignore(TypeError, if_else=aviolation(context)):
            await async_database.atest_hmac(atag, str(ainputs))

        # async hmac tags fail when inputs are altered
        aiinputs = int.from_bytes(ainputs, BYTE_ORDER)
        context = "Async data value alteration not caught!"
        for bit in range(aiinputs.bit_length()):
            async with aignore(async_database.InvalidHMAC, if_else=aviolation(context)):
                await async_database.atest_hmac(atag, (aiinputs ^ (1 << bit)).to_bytes(32, BYTE_ORDER))

        # async hmac tags fail when they are altered
        aitag = int.from_bytes(atag, BYTE_ORDER)
        context = "Async tag alteration not caught!"
        for abit in range(aitag.bit_length()):
            async with aignore(async_database.InvalidHMAC, if_else=aviolation(context)):
                altered_tag = (aitag ^ (1 << abit)).to_bytes(32, BYTE_ORDER)
                await async_database.atest_hmac(altered_tag, ainputs)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *


ainvalid_size_datastream = adata(plaintext_bytes, size=BLOCKSIZE + 1)
invalid_size_datastream = data(plaintext_bytes, size=BLOCKSIZE + 1)


async def test_datastream_limits():
    akey_bundle = await KeyAADBundle(key=key, salt=salt, aad=aad, allow_dangerous_determinism=True).async_mode()
    key_bundle = KeyAADBundle(key=key, salt=salt, aad=aad, allow_dangerous_determinism=True).sync_mode()

    # async cipher blocksize limits are respected
    context = f"A cipher block exceeded {BLOCKSIZE} bytes!"
    async with aignore(OverflowError, if_else=aviolation(context)):
        await ainvalid_size_datastream.areset()
        keystream = abytes_keys.root(akey_bundle)
        shmac = StreamHMAC(akey_bundle)._for_encryption()
        async for chunk in abytes_encipher(ainvalid_size_datastream, shmac):
            pass

    # sync cipher blocksize limits are respected
    context = f"A cipher block exceeded {BLOCKSIZE} bytes"
    with ignore(OverflowError, if_else=violation(context)):
        invalid_size_datastream.reset()
        keystream = bytes_keys(key_bundle)
        shmac = StreamHMAC(key_bundle)._for_encryption()
        for chunk in bytes_encipher(invalid_size_datastream, shmac):
            pass


def test_keys_limits():
    context = "A falsey key was not overwritten"
    with ignore(AssertionError, if_else=violation(context)):
        key_bundle = KeyAADBundle(key=None)
        assert not key_bundle.key

    context = "Non-bytes key was allowed"
    with ignore(TypeError, if_else=violation(context)):
        KeyAADBundle(key=csprng().hex(), allow_dangerous_determinism=True)


def test_salt_limits():
    context = "Non-bytes salt was allowed"
    with ignore(TypeError, if_else=violation(context)):
        KeyAADBundle(salt=csprng().hex())

    context = "Invalid length salt was allowed"
    with ignore(ValueError, if_else=violation(context)):
        KeyAADBundle(salt=csprng(), allow_dangerous_determinism=True)


async def test_salt_reuse_resistance_given_by_ivs_only():
    number_of_tests = 256
    unpadded_plaintext = BLOCKSIZE * b"\x00"

    # ASYNC
    # aggregate the first ciphertext block of a collection of
    # async ciphertexts instantiated with the same key, salt & aad
    kw = dict(key=key, salt=salt, allow_dangerous_determinism=True)
    aciphertexts = set({
        await aunpack(
            abytes_encipher(
                adata.root(unpadded_plaintext),
                shmac=StreamHMAC(await KeyAADBundle(**kw).async_mode())._for_encryption(),
            )
        ).asend(None)
        for _ in range(number_of_tests)
    })
    assert all(len(block) == BLOCKSIZE for block in aciphertexts)

    # the vulnerable first block of async ciphertexts is always
    # unique
    assert len(aciphertexts) == number_of_tests

    # the most vulnerable first INNER_HEADER-bytes of async
    # ciphertexts are also always unique
    ainner_headers = {aciphertext[INNER_HEADER_SLICE] for aciphertext in aciphertexts}
    assert len(ainner_headers) == number_of_tests


    # SYNC
    # aggregate the first ciphertext block of a collection of
    # ciphertexts instantiated with the same key, salt & aad
    kw = dict(key=key, salt=salt, allow_dangerous_determinism=True)
    ciphertexts = set({
        unpack(
            bytes_encipher(
                data.root(unpadded_plaintext),
                shmac=StreamHMAC(KeyAADBundle(**kw).sync_mode())._for_encryption(),
            )
        ).send(None)
        for _ in range(number_of_tests)
    })
    assert all(len(block) == BLOCKSIZE for block in ciphertexts)

    # the vulnerable first block of ciphertexts is always unique
    assert len(ciphertexts) == number_of_tests

    # the most vulnerable first INNER_HEADER-bytes of ciphertexts are
    # also always unique
    inner_headers = {ciphertext[INNER_HEADER_SLICE] for ciphertext in ciphertexts}
    assert len(inner_headers) == number_of_tests


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})

